import pytest  # type: ignore
import json
from pytos2.securechange.designer import (
    DesignerResults,
    DeviceSuggestion,
    BindingSuggestion,
    get_instruction_type,
    classify_object_type,
    classify_service_type,
    NetworkObject,
    HostObject,
    RangeObject,
    SubnetObject,
    GroupObject,
    Instruction,
    AddNetworkObjectInstruction,
    UpdateRuleInstruction,
    AddNewRuleInstruction,
    FullyImplementedInstruction,
    ServiceObject,
    IPService,
    ServiceGroup,
    Rule,
)

from pytos2.securechange.designer_verifier_common import (
    TransportService,
    ICMPService,
)

from pytos2.utils import get_api_node

import responses


# need to load proper json file
@pytest.fixture
def load_instruction_json_by_type():
    def f(instruction_class):
        return instruction_class.kwargify(
            json.load(
                open(
                    f"tests/securechange/json/field/{instruction_class._type.value}-field.json"
                )
            )
        )


class TestDesignerResults:
    @pytest.fixture
    def designer_results(self):
        j = json.load(open("tests/securechange/json/designer/designer_results.json"))
        return DesignerResults.kwargify(j)

    def test_attributes(self, designer_results):
        assert designer_results.id == 1458
        assert hasattr(designer_results, "device_suggestions")
        assert isinstance(designer_results, DesignerResults)

    def test_get_instruction_by_id(self, designer_results):
        instruction_2967 = designer_results.get_instruction_by_id(2967)
        assert instruction_2967.id == 2967
        assert instruction_2967.instruction_type == Instruction.InstructionType.NEW_RULE
        assert isinstance(instruction_2967, Instruction)
        assert isinstance(instruction_2967, AddNewRuleInstruction)


class TestDeviceSuggestion:
    @pytest.fixture
    def suggestion(self):
        j = json.load(open("tests/securechange/json/designer/designer_results.json"))
        device_suggestion_node = get_api_node(
            j, "designer_results.suggestions_per_device.device_suggestion"
        )
        return DeviceSuggestion.kwargify(device_suggestion_node)

    def test_attributes(self, suggestion):
        assert suggestion.id == 815
        assert suggestion.management_name == "DG-B"
        assert suggestion.vendor_name == "PaloAltoNetworks"
        assert suggestion.offline_device is False
        assert suggestion.ancestor_management_name == "PANO-8"


class TestBindingSuggestion:
    @pytest.fixture
    def suggestion(self):
        j = json.load(open("tests/securechange/json/designer/designer_results.json"))
        binding_suggestion_node = get_api_node(
            j,
            "designer_results.suggestions_per_device.device_suggestion.suggestions_per_binding.binding_suggestion",
        )
        return BindingSuggestion.kwargify(binding_suggestion_node)

    def test_attributes(self, suggestion):
        assert suggestion.binding_uid == "{00-FACED0-00}"
        assert type(suggestion.binding_name) is str


class TestFullyImplementedInstruction:
    @pytest.fixture
    def instruction(self):
        j = json.load(
            open("tests/securechange/json/designer/fully_implemented-instruction.json")
        )
        return get_instruction_type(j)

    def test_attributes(self, instruction):
        assert instruction.implements_access_requests == ["AR1"]
        assert instruction.status == "DESIGN_FULLY_IMPLEMENTED"
        assert (
            instruction.instruction_type
            == Instruction.InstructionType.FULLY_IMPLEMENTED
        )
        assert hasattr(instruction, "device_added_service_object") is False
        assert hasattr(instruction, "device_added_network_object") is False
        assert hasattr(instruction, "object") is False
        assert hasattr(instruction, "rule") is False or instruction.rule is None

    def test_object_inheritance(self, instruction):
        assert isinstance(instruction, FullyImplementedInstruction)
        assert isinstance(instruction, Instruction)


class TestFullyImplementedNoSecurityInstruction:
    @pytest.fixture
    def instruction(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/fully_implemented_since_no_security-instruction.json"
            )
        )
        return get_instruction_type(j)

    def test_attributes(self, instruction):
        assert instruction.implements_access_requests == ["AR1"]
        assert instruction.status == "DESIGN_FULLY_IMPLEMENTED"
        assert (
            instruction.instruction_type
            == Instruction.InstructionType.NO_SECURITY_ON_INTERFACE
        )
        assert hasattr(instruction, "device_added_service_object") is False
        assert hasattr(instruction, "device_added_network_object") is False
        assert hasattr(instruction, "object") is False
        assert hasattr(instruction, "rule") is False or instruction.rule is None

    def test_object_inheritance(self, instruction):
        assert isinstance(instruction, FullyImplementedInstruction)
        assert isinstance(instruction, Instruction)


class TestAddHostObjectInstruction:
    @pytest.fixture
    def instruction(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_network_host_object-instruction.json"
            )
        )
        return get_instruction_type(j)

    def test_attributes(self, instruction):
        assert instruction.implements_access_requests == ["AR1"]
        assert instruction.status == "DESIGN_SUCCESS"
        assert (
            instruction.instruction_type
            == Instruction.InstructionType.ADD_OBJECT_TO_DEVICE
        )
        assert hasattr(instruction, "device_added_service_object") is False
        assert hasattr(instruction, "device_added_network_object") is False
        assert hasattr(instruction, "object") is True
        # assert isinstance(instruction.modified_object_name, str)  # not mapped

    def test_object_inheritance(self, instruction):
        assert isinstance(instruction, AddNetworkObjectInstruction)
        assert isinstance(instruction, Instruction)


class TestAddedHostObject:
    @pytest.fixture
    def device_added_host_object(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_network_host_object-instruction.json"
            )
        )
        device_added_host_object = get_api_node(j, "device_added_network_object")
        return classify_object_type(device_added_host_object)

    def test_attributes(self, device_added_host_object):
        assert device_added_host_object.uid == "-9"
        assert device_added_host_object.name == "Host_172.16.90.5"
        assert device_added_host_object.display_name == "Host_172.16.90.5"
        assert device_added_host_object.origin == "DEVICE"
        assert device_added_host_object.is_global is False
        assert device_added_host_object.implicit is False
        assert device_added_host_object.shared is False
        assert device_added_host_object.id == 3156
        assert device_added_host_object.ip_type == NetworkObject.IPType.IPV4

    def test_object_inheritance(self, device_added_host_object):
        assert isinstance(device_added_host_object, NetworkObject)
        assert isinstance(device_added_host_object, HostObject)


class TestAddRangeObjectInstruction:
    @pytest.fixture
    def instruction(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_network_range_object-instruction.json"
            )
        )
        return get_instruction_type(j)

    def test_attributes(self, instruction):
        assert instruction.implements_access_requests == ["AR1"]
        assert instruction.status == "DESIGN_SUCCESS"
        assert (
            instruction.instruction_type
            == Instruction.InstructionType.ADD_OBJECT_TO_DEVICE
        )
        assert hasattr(instruction, "device_added_service_object") is False
        assert hasattr(instruction, "device_added_network_object") is False
        assert hasattr(instruction, "object") is True

    def test_object_inheritance(self, instruction):
        assert isinstance(instruction, AddNetworkObjectInstruction)
        assert isinstance(instruction, Instruction)


class TestAddedRangeObject:
    @pytest.fixture
    def device_added_range_object(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_network_range_object-instruction.json"
            )
        )
        device_added_range_object = get_api_node(j, "device_added_network_object")
        return classify_object_type(device_added_range_object)

    def test_attributes(self, device_added_range_object):
        assert device_added_range_object.uid == "-13"
        assert device_added_range_object.name == "1.2.3.4-1.2.3.10"
        assert device_added_range_object.display_name == "1.2.3.4-1.2.3.10"
        assert device_added_range_object.origin == "DEVICE"
        assert device_added_range_object.is_global is False
        assert device_added_range_object.implicit is True
        assert device_added_range_object.shared is False
        assert device_added_range_object.id == 3290
        assert device_added_range_object.ip_type == NetworkObject.IPType.IPV4
        assert device_added_range_object.min_ip == "1.2.3.4"
        assert device_added_range_object.max_ip == "1.2.3.10"

    def test_object_inheritance(self, device_added_range_object):
        assert isinstance(device_added_range_object, NetworkObject)
        assert isinstance(device_added_range_object, RangeObject)


class TestAddSubnetObjectInstruction:
    @pytest.fixture
    def instruction(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_network_subnet_object-instruction.json"
            )
        )
        return get_instruction_type(j)

    def test_attributes(self, instruction):
        assert instruction.implements_access_requests == ["AR1"]
        assert instruction.status == "DESIGN_SUCCESS"
        assert (
            instruction.instruction_type
            == Instruction.InstructionType.ADD_OBJECT_TO_DEVICE
        )
        assert hasattr(instruction, "device_added_service_object") is False
        assert hasattr(instruction, "device_added_network_object") is False
        assert hasattr(instruction, "object") is True

    def test_object_inheritance(self, instruction):
        assert isinstance(instruction, AddNetworkObjectInstruction)
        assert isinstance(instruction, Instruction)


class TestAddedSubnetObject:
    @pytest.fixture
    def device_added_subnet_object(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_network_subnet_object-instruction.json"
            )
        )
        device_added_subnet_object = get_api_node(j, "device_added_network_object")
        return classify_object_type(device_added_subnet_object)

    def test_attributes(self, device_added_subnet_object):
        assert device_added_subnet_object.uid == "-1"
        assert device_added_subnet_object.name == "Subnet_9.8.7.0"
        assert device_added_subnet_object.display_name == "Subnet_9.8.7.0"
        assert device_added_subnet_object.origin == "DEVICE"
        assert device_added_subnet_object.is_global is False
        assert device_added_subnet_object.implicit is False
        assert device_added_subnet_object.shared is False
        assert device_added_subnet_object.id == 3305
        assert device_added_subnet_object.ip_type == NetworkObject.IPType.IPV4

    def test_object_inheritance(self, device_added_subnet_object):
        assert isinstance(device_added_subnet_object, NetworkObject)
        assert isinstance(device_added_subnet_object, SubnetObject)


class TestAddGroupObjectInstruction:
    @pytest.fixture
    def instruction(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_network_subnet_object-instruction.json"
            )
        )
        return get_instruction_type(j)

    def test_attributes(self, instruction):
        assert instruction.implements_access_requests == ["AR1"]
        assert instruction.status == "DESIGN_SUCCESS"
        assert (
            instruction.instruction_type
            == Instruction.InstructionType.ADD_OBJECT_TO_DEVICE
        )
        assert hasattr(instruction, "device_added_service_object") is False
        assert hasattr(instruction, "device_added_network_object") is False
        assert hasattr(instruction, "object") is True

    def test_object_inheritance(self, instruction):
        assert isinstance(instruction, AddNetworkObjectInstruction)
        assert isinstance(instruction, Instruction)


class TestAddedGroupObject:
    @pytest.fixture
    def device_added_group_object(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_network_group_object-instruction.json"
            )
        )
        device_added_group_object = get_api_node(j, "device_added_network_object")
        return classify_object_type(device_added_group_object)

    def test_attributes(self, device_added_group_object):
        assert device_added_group_object.uid == "-17"
        assert device_added_group_object.name == "NetworkGroup_45"
        assert device_added_group_object.display_name == "NetworkGroup_45"
        assert device_added_group_object.origin == "DEVICE"
        assert device_added_group_object.is_global is False
        assert device_added_group_object.implicit is False
        assert device_added_group_object.shared is False
        assert device_added_group_object.id == 3292
        assert device_added_group_object.ip_type == NetworkObject.IPType.OTHER
        assert hasattr(device_added_group_object, "members")
        assert isinstance(device_added_group_object.members, list)

    def test_object_inheritance(self, device_added_group_object):
        assert isinstance(device_added_group_object, NetworkObject)
        assert isinstance(device_added_group_object, GroupObject)

    def test_member_attribute(self, device_added_group_object):
        assert device_added_group_object.members[0].uid == "-13"
        assert device_added_group_object.members[0].name == "1.2.3.4-1.2.3.10"
        assert device_added_group_object.members[0].display_name == "1.2.3.4-1.2.3.10"
        assert device_added_group_object.members[0].origin == "DEVICE"
        assert device_added_group_object.members[0].is_global is False
        assert device_added_group_object.members[0].implicit is True
        assert device_added_group_object.members[0].shared is False
        assert device_added_group_object.members[0].id == 3290
        assert device_added_group_object.members[0].ip_type == NetworkObject.IPType.IPV4
        assert device_added_group_object.members[0].min_ip == "1.2.3.4"
        assert device_added_group_object.members[0].max_ip == "1.2.3.10"

    def test_member_inheritance(self, device_added_group_object):
        assert isinstance(device_added_group_object.members[0], NetworkObject)
        assert isinstance(device_added_group_object.members[0], RangeObject)


class TestAddTransportServiceObject:
    @pytest.fixture
    def device_added_service_object(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_transport_service_object-instruction.json"
            )
        )
        device_added_service_object = get_api_node(j, "device_added_service_object")
        return classify_service_type(device_added_service_object)

    def test_attributes(self, device_added_service_object):
        assert device_added_service_object.uid == "-21"
        assert device_added_service_object.name == "TCP_1234"
        assert device_added_service_object.display_name == "TCP_1234"
        assert device_added_service_object.origin == "DEVICE"
        assert device_added_service_object.is_global is False
        assert device_added_service_object.implicit is False
        assert device_added_service_object.shared is False
        assert device_added_service_object.id == 1361
        assert device_added_service_object.protocol == 6
        assert device_added_service_object.min_port == 1234
        assert device_added_service_object.max_port == 1234


class TestAddICMPServiceObject:
    @pytest.fixture
    def device_added_service_object(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_icmp_service_object-instruction.json"
            )
        )
        device_added_service_object = get_api_node(j, "device_added_service_object")
        return classify_service_type(device_added_service_object)

    def test_attributes(self, device_added_service_object):
        assert device_added_service_object.uid == "-15"
        assert device_added_service_object.name == "ICMP_253"
        assert device_added_service_object.display_name == "ICMP_253"
        assert device_added_service_object.origin == "DEVICE"
        assert device_added_service_object.is_global is False
        assert device_added_service_object.implicit is False
        assert device_added_service_object.shared is False
        assert device_added_service_object.id == 1435
        assert hasattr(device_added_service_object, "protocol") is False
        assert hasattr(device_added_service_object, "min_port") is False
        assert hasattr(device_added_service_object, "max_port") is False
        assert device_added_service_object.min_icmp_type == 253
        assert device_added_service_object.max_icmp_type == 253


class TestAddServiceGroup:
    @pytest.fixture
    def device_added_service_group(self):
        j = json.load(
            open("tests/securechange/json/designer/add_service_group-instruction.json")
        )
        device_added_service_group = get_api_node(j, "device_added_service_object")
        return classify_service_type(device_added_service_group)

    def test_attributes(self, device_added_service_group):
        assert device_added_service_group.uid == "-33"
        assert device_added_service_group.name == "ServiceGroup_32"
        assert device_added_service_group.display_name == "ServiceGroup_32"
        assert device_added_service_group.origin == "DEVICE"
        assert device_added_service_group.is_global is False
        assert device_added_service_group.implicit is False
        assert device_added_service_group.shared is False
        assert device_added_service_group.id == 1432
        assert hasattr(device_added_service_group, "protocol") is False
        assert hasattr(device_added_service_group, "min_port") is False
        assert hasattr(device_added_service_group, "max_port") is False
        assert hasattr(device_added_service_group, "min_icmp_type") is False
        assert hasattr(device_added_service_group, "max_icmp_type") is False
        assert isinstance(device_added_service_group.members, list)

    def test_object_inheritance(self, device_added_service_group):
        assert isinstance(device_added_service_group, ServiceObject)
        assert isinstance(device_added_service_group, ServiceGroup)

    def test_member_attribute(self, device_added_service_group):
        assert device_added_service_group.members[0].uid == "-29"
        assert device_added_service_group.members[0].display_name == "icmp 253"
        assert device_added_service_group.members[0].origin == "DEVICE"
        assert device_added_service_group.members[0].is_global is False
        assert device_added_service_group.members[0].implicit is True
        assert device_added_service_group.members[0].shared is False
        assert device_added_service_group.members[0].id == 1433
        assert device_added_service_group.members[0].min_icmp_type == 253
        assert device_added_service_group.members[0].max_icmp_type == 253

        assert device_added_service_group.members[1].uid == "-31"
        assert device_added_service_group.members[1].display_name == "udp 343"
        assert device_added_service_group.members[1].origin == "DEVICE"
        assert device_added_service_group.members[1].is_global is False
        assert device_added_service_group.members[1].implicit is True
        assert device_added_service_group.members[1].shared is False
        assert device_added_service_group.members[1].id == 1434
        assert device_added_service_group.members[1].protocol == 17
        assert device_added_service_group.members[1].min_port == 343
        assert device_added_service_group.members[1].max_port == 343


class TestAddNewRuleInstruction:
    @pytest.fixture
    def instruction(self):
        j = json.load(
            open("tests/securechange/json/designer/add_new_rule-instruction.json")
        )
        return get_instruction_type(j)

    def test_attributes(self, instruction):
        assert instruction.implements_access_requests == ["AR1"]
        assert instruction.status == "DESIGN_SUCCESS"
        assert instruction.instruction_type == Instruction.InstructionType.NEW_RULE
        assert instruction.rule_placement.value == "BEFORE"
        assert instruction.change_action.value == "ADD"
        assert hasattr(instruction, "device_added_service_object") is False
        assert hasattr(instruction, "device_added_network_object") is False
        assert hasattr(instruction, "object") is False
        assert hasattr(instruction, "rule") is True

    def test_object_inheritance(self, instruction):
        assert isinstance(instruction, AddNewRuleInstruction)
        assert isinstance(instruction, Instruction)


class TestNewRule:
    @pytest.fixture
    def rule(self):
        j = json.load(
            open("tests/securechange/json/designer/add_new_rule-instruction.json")
        )
        rule_json = get_api_node(j, "rule")
        return Rule.kwargify(rule_json)

    def test_attributes(self, rule):
        assert rule.uid == "-24"
        assert rule.is_disabled is False
        assert rule.from_zone == "Inside"
        assert rule.to_zone == "DMZ"

    def test_object_inheritance(self, rule):
        assert isinstance(rule, Rule)


class TestNewRuleAnyService:
    @pytest.fixture
    def rule(self):
        j = json.load(
            open(
                "tests/securechange/json/designer/add_new_rule_any_service-instruction.json"
            )
        )
        rule_json = get_api_node(j, "rule")
        return Rule.kwargify(rule_json)

    def test_attributes(self, rule):
        assert rule.uid == "-22"
        assert rule.is_disabled is False
        assert rule.from_zone == "any"
        assert rule.to_zone == "any"

    def test_object_inheritance(self, rule):
        assert isinstance(rule, Rule)

    def test_any_service(self, rule):
        assert isinstance(rule.services, list)
        assert rule.services[0].uid == "00000000-0000-0000-0000-000000000000"
        assert rule.services[0].name == "Any"
        assert rule.services[0].origin == "PROVISIONING"
        assert rule.services[0].is_global is False
        assert rule.services[0].implicit is False
        assert rule.services[0].shared is False
        assert rule.services[0].min_protocol == 0
        assert rule.services[0].max_protocol == 255
        assert isinstance(rule.services[0], IPService)


class TestUpdateRuleInstruction:
    @pytest.fixture
    def instruction(self):
        j = json.load(
            open("tests/securechange/json/designer/update_rule-instruction.json")
        )
        return get_instruction_type(j)

    def test_attributes(self, instruction):
        assert instruction.implements_access_requests == ["AR1"]
        assert instruction.status == "DESIGN_SUCCESS"
        assert instruction.instruction_type == Instruction.InstructionType.UPDATE_RULE
        assert hasattr(instruction, "device_added_service_object") is False
        assert hasattr(instruction, "device_added_network_object") is False
        assert hasattr(instruction, "object") is False
        assert hasattr(instruction, "rule") is True
        assert (
            instruction.rule.destination_services[0].referenced
            == ServiceObject.Referenced.UNKNOWN
        )

        assert len(instruction.sources) > 0
        assert len(instruction.destinations) > 0
        assert len(instruction.services) > 0
        assert instruction.source_objects is instruction.sources

    def test_object_inheritance(self, instruction):
        assert isinstance(instruction, UpdateRuleInstruction)
        assert isinstance(instruction, Instruction)


@pytest.fixture
def services_mock(st, st_devices_mock):
    responses.add(
        responses.GET,
        "https://198.18.0.1/securetrack/api/devices/1/services",
        json=json.load(open("tests/securetrack/json/services/1.json")),
    )


class TestServiceObjectCasts:
    @responses.activate
    def test_casts(self, st, services_mock):
        services = st.get_services(1)
        assert services is not None

        svc0 = services[0]
        svc = ServiceObject.from_securetrack(svc0)
        assert svc.min_port == 902
        assert svc.max_port == 902
        assert isinstance(svc, TransportService)

        svc419 = services[419]
        svc = ServiceObject.from_securetrack(svc419)
        assert svc.management_domain == "Shared"
        assert isinstance(svc, ICMPService)



import pytest
import json
import responses

from pytos2.securetrack.cloud import JoinCloud
from pytos2.utils import get_api_node

from netaddr import IPAddress


class TestCloud:
    @responses.activate
    def test_get_clouds_200(self, st):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/topology/clouds",
            json=json.load(
                open("tests/securetrack/json/clouds/all_non_joined_clouds.json")
            ),
            status=200,
        )

        clouds = st.get_clouds()
        cl = clouds[0]
        assert cl.id == 5
        assert cl.name == "Cloud 192.168.111.1"
        assert cl.domain == 1
        assert cl.type.value == "NON_JOINED"
        assert cl.ip == IPAddress("192.168.111.1")
        assert isinstance(cl.members, list)

    @responses.activate
    def test_get_cloud(self, st):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/topology/clouds/4",
            json=json.load(open("tests/securetrack/json/clouds/cloud_4.json")),
            status=200,
        )

        cl = st.get_cloud(4)
        assert cl.id == 4
        assert cl.name == "Cloud 10.100.100.19"
        assert cl.domain == 1
        assert cl.type.value == "NON_JOINED"
        assert cl.ip == IPAddress("10.100.100.19")
        assert isinstance(cl.members, list)

    @responses.activate
    def test_get_cloud_internal_networks(self, st):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/topology/cloud_internal_networks/58",
            json=json.load(
                open("tests/securetrack/json/clouds/cloud_internal_networks_58.json")
            ),
            status=200,
        )

        networks = st.get_cloud_internal_networks(58)
        ntwk = networks[0]
        assert ntwk.ip == IPAddress("192.168.74.1")
        assert ntwk.mask == IPAddress("255.255.255.0")

    @responses.activate
    def test_add_topology_cloud(self, st):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/clouds",
            status=200,
        )
        response = st.add_topology_cloud(cloud_name="Cloud 888", cloud_members=[49, 51])
        assert response is None

    @responses.activate
    def test_update_topology_cloud(self, st):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/clouds/105",
            status=200,
        )

        response = st.update_topology_cloud(105, cloud_name="Cloud 888")
        assert response is None

        response = st.update_topology_cloud(105, cloud_members=[52])
        assert response is None

        response = st.update_topology_cloud(105, cloud_members=[52], action="remove")
        assert response is None

    @responses.activate
    def test_get_cloud_suggestions(self, st):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/topology/cloud_suggestions",
            json=json.load(
                open("tests/securetrack/json/clouds/cloud_suggestions.json")
            ),
            status=200,
        )

        cloud_suggestions = st.get_cloud_suggestions()
        cl = cloud_suggestions[0]
        assert cl.management_name == "ACI Fabric"
        assert cl.management_id == 234
        assert cl.cloud_name == "Cloud 10.30.27.6"
        assert cl.cloud_id == 32
        assert cl.vertex_id == 485
        assert cl.ip == IPAddress("10.30.27.6")
        assert cl.routes_count == 134
        assert cl.is_parent is False

    @responses.activate
    def test_get_cloud_suggestions_by_id(self, st):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securetrack/api/topology/cloud_suggestions/32",
            json=json.load(
                open("tests/securetrack/json/clouds/cloud_suggestions_32.json")
            ),
            status=200,
        )

        # This somewhat unexpectedly returns a list
        cl = st.get_cloud_suggestions_by_id(32)[0]
        assert cl.management_name == "ACI Fabric"
        assert cl.management_id == 234
        assert cl.cloud_name == "Cloud 10.30.27.6"
        assert cl.cloud_id == 32
        assert cl.vertex_id == 485
        assert cl.ip == IPAddress("10.30.27.6")
        assert cl.routes_count == 134
        assert cl.is_parent is False


class TestJoinCloud:
    cloud = json.load(open("tests/securetrack/json/join_clouds/cloud-67.json"))

    @responses.activate
    def test_add_join_cloud_200(self, st):
        """Add cloud"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=201,
        )

        """Add single cloud"""
        newCloud = st.add_join_cloud(
            name=self.cloud["name"], clouds=self.cloud["clouds"]
        )
        assert newCloud is None

    @responses.activate
    def test_add_join_cloud_400(self, st):
        """Add cloud"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=400,
        )

        """Add single cloud"""
        with pytest.raises(ValueError) as exception:
            st.add_join_cloud(name=self.cloud["name"], clouds=self.cloud["clouds"])
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_join_cloud_404(self, st):
        """Add cloud"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=404,
        )

        """Add single cloud"""
        with pytest.raises(ValueError) as exception:
            st.add_join_cloud(name=self.cloud["name"], clouds=self.cloud["clouds"])
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_join_cloud(self, st, join_cloud_mock):
        """Get cloud by device id"""
        join_cloud = JoinCloud.kwargify(self.cloud)
        cloudByInt = st.get_join_cloud(67)
        cloudByStr = st.get_join_cloud("67")
        assert cloudByInt == join_cloud
        assert cloudByStr == join_cloud

        with pytest.raises(ValueError) as exception:
            st.get_join_cloud(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_join_cloud("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_update_join_cloud_200(self, st):
        """Update cloud"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=200,
        )

        """Update single cloud"""
        cloud_to_update = JoinCloud.kwargify(self.cloud)
        newCloud = st.update_join_cloud(cloud_to_update)
        assert newCloud is None

    @responses.activate
    def test_update_join_cloud_400(self, st):
        """PUT bad request"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            cloud_to_update = JoinCloud.kwargify(self.cloud)
            st.update_join_cloud(cloud_to_update)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_update_join_cloud_404(self, st):
        """PUT device id not found"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/join/clouds",
            status=404,
        )

        with pytest.raises(ValueError) as exception:
            cloud_to_update = JoinCloud.kwargify(self.cloud)
            st.update_join_cloud(cloud_to_update)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_join_cloud(self, st, join_cloud_mock):
        """Delete cloud by cloud id"""
        cloudByInt = st.delete_join_cloud(67)
        cloudByStr = st.delete_join_cloud("67")
        assert cloudByInt is None
        assert cloudByStr is None

        with pytest.raises(ValueError) as exception:
            st.delete_join_cloud(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.delete_join_cloud("404")
        assert "Not Found" in str(exception.value)



import pytest
import json

from netaddr import IPAddress, IPNetwork
from pytos2.securetrack.interface import Interface, BindableObject, TopologyInterface
from pytos2.utils import get_api_node


class TestInterface:
    @pytest.fixture
    def interface(self):
        j = json.load(open("tests/securetrack/json/interfaces/device8-interfaces.json"))
        interface_node = get_api_node(j, "interfaces.interface")[0]
        return Interface.kwargify(interface_node)


class TestBindableObject:
    @pytest.fixture
    def bindable(self):
        j = json.load(open("tests/securetrack/json/interfaces/device8-objects.json"))
        object_node = get_api_node(j, "bindable_objects")[1]
        return BindableObject.kwargify(object_node)


class TestTopologyInterface:
    @pytest.fixture
    def topointerface(self):
        j = json.load(
            open("tests/securetrack/json/interfaces/device8-topology-interfaces.json")
        )
        object_node = get_api_node(j, "interface")[0]
        return TopologyInterface.kwargify(object_node)

    def test_attributes(self, topointerface):
        assert topointerface.name == "Datacenter"
        assert topointerface.device_id == 8
        assert topointerface.ip.prefixlen == 24
        assert topointerface.ip.ip == IPAddress("10.3.3.1")
        assert isinstance(topointerface.ip, IPNetwork)


class TestAzureTopologyInterface:
    @pytest.fixture
    def topointerface(self):
        j = json.load(
            open(
                "tests/securetrack/json/interfaces/device784-azure-topology-interfaces.json"
            )
        )
        object_node = get_api_node(j, "interface")[3]
        return TopologyInterface.kwargify(object_node)

    def test_attributes(self, topointerface):
        assert (
            topointerface.name
            == "/subscriptions/44444444-xxxx-yyyy-zzzz-555555555555/resourceGroups/QA-RG/providers/Microsoft.Network/virtualHubs/Main-Hub-WE"
        )
        assert topointerface.device_id == 784
        assert topointerface.ip.prefixlen == 24
        assert topointerface.ip.ip == IPAddress("10.0.10.1")
        assert isinstance(topointerface.ip, IPNetwork)



import pytest
import json
import responses
from netaddr import IPAddress

from pytos2.securechange.extension import MarketplaceApp
from pytos2.utils import get_api_node


class TestExtensions:
    @responses.activate
    def test_get_extensions(self, extensions_mock, scw):
        exts = scw.get_extensions()
        assert exts[2].app_id == "ipam"
        assert exts[2].name == "IPAM Security Policy App"
        assert exts[2].url == "/apps/ispa"
        assert (
            exts[2].description
            == "The Tufin IPAM sync app integrates Tufin SecureTrack with IP Address Management systems to populate networks zones in order to use the USP in SecureTrack."
        )
        assert exts[2].version == "4.1.5"
        assert exts[2].installed is True
        assert exts[2].upsell is False



from enum import Enum
from typing import Optional, List, ForwardRef

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, multiroot_kwargify


class WorkflowType(Enum):
    GENERIC = "GENERIC"
    ACCESS_REQUEST_AND_MODIFY_GROUP = "ACCESS_REQUEST_AND_MODIFY_GROUP"
    ACCESS_REQUEST = "ACCESS_REQUEST"
    MODIFY_GROUP = "MODIFY_GROUP"
    DECOMMISSION_NETWORK_OBJECT = "DECOMMISSION_NETWORK_OBJECT"
    RULE_DECOMMISSION = "RULE_DECOMMISSION"
    RULE_RECERTIFICATION = "RULE_RECERTIFICATION"
    CLONE_NETWORK_OBJECT_POLICY = "CLONE_NETWORK_OBJECT_POLICY"
    RULE_MODIFICATION = "RULE_MODIFICATION"


@propify
class BasicWorkflowInfo(Jsonable):
    id: Optional[int] = prop(None)
    description: Optional[str] = prop("")
    name: Optional[str] = prop("")
    type: Optional[WorkflowType] = prop("")


@propify
class WorkflowFieldConfiguration(Jsonable):
    pass


@propify
class TicketTaskField(Jsonable):
    pass


@propify
class WorkflowField(Jsonable):
    class Prop(Enum):
        FIELD_VALUE = "fieldValue"

    name: Optional[str] = prop(None)
    tooltip: Optional[str] = prop(None)
    mandatory: Optional[bool] = prop(None)
    multiple: Optional[bool] = prop(None)
    type: Optional[str] = prop(None)
    configuration: Optional[WorkflowFieldConfiguration] = prop(None)
    field_value: Optional[TicketTaskField] = prop(None, key=Prop.FIELD_VALUE)

    pass


@propify
class WorkflowStep(Jsonable):
    pass


@propify
class FullWorkflow(Jsonable):
    class Config:
        has_multi_root = True
        root_default_cls = "FullWorkflow"
        root_dict = {"access_request_workflow": "FullARWorkflow"}

    class Prop(Enum):
        IS_ACTIVE = "isActive"
        IS_VALID = "isValid"
        REFERENCE_TICKET_ENABLED = "referenceTicketEnabled"
        DESCRIPTION = "description"
        ALLOW_REQUESTER_CONFIRMATION = "allowRequesterConfirmation"
        TIME_UNIT_AUTO_CONFIRM = "timeUnitAutoConfirm"
        TIME_TO_AUTO_CONFIRM = "timeToAutoConfirm"
        REOPEN_STEP_NAME = "reopenStepName"
        CLOSE_IMPLEMENTED = "closeImplemented"
        GLOBAL_FIELDS = "globalFields"

    id: Optional[int] = prop(None)
    type: Optional[WorkflowType] = prop(WorkflowType.GENERIC)
    name: Optional[str] = prop(None)
    is_active: Optional[bool] = prop(None, key=Prop.IS_ACTIVE)
    is_valid: Optional[bool] = prop(None, key=Prop.IS_VALID)
    reference_ticket_enabled: Optional[bool] = prop(
        None, key=Prop.REFERENCE_TICKET_ENABLED
    )
    description: Optional[str] = prop(None, key=Prop.DESCRIPTION)
    allow_requester_confirmation: Optional[bool] = prop(
        None, key=Prop.ALLOW_REQUESTER_CONFIRMATION
    )
    time_unit_auto_confirm: Optional[str] = prop(None, key=Prop.TIME_UNIT_AUTO_CONFIRM)
    time_to_auto_confirm: Optional[int] = prop(None, key=Prop.TIME_TO_AUTO_CONFIRM)
    reopen_step_name: Optional[str] = prop(None, key=Prop.REOPEN_STEP_NAME)
    close_implemented: Optional[bool] = prop(None, key=Prop.CLOSE_IMPLEMENTED)
    steps: List[WorkflowStep] = prop(factory=list)
    global_fields: List[WorkflowField] = prop(factory=list)

    @classmethod
    def kwargify(cls, data):
        return multiroot_kwargify(cls, data, _globals=globals(), _locals=locals())


@propify
class FullARWorkflow(FullWorkflow):
    pass



from typing import List, Optional, Union
from enum import Enum
from datetime import datetime
from pytos2.models import Jsonable, UnMapped
from pytos2.utils import prop, propify, safe_iso8601_date


@propify
class RequestsSearchList(Jsonable):
    class Meta(Enum):
        ROOT = "requests"

    requests: List["RequestSearch"] = prop(factory=list)


@propify
class RequestSearch(Jsonable):
    class TicketType(Enum):
        TICKET = "TICKET"
        TICKET_DRAFT = "TICKET_DRAFT"

    class Prop(Enum):
        CREATE_DATE = "create_date"
        UPDATE_DATE = "update_date"

    id: Optional[int] = prop(None)
    type: Optional[TicketType] = prop(None, kwargify=str)  # TICKET or TICKET_DRAFT
    priority: Optional[str] = prop(None)
    create_date: Optional[datetime] = prop(
        None,
        key=Prop.CREATE_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    update_date: Optional[datetime] = prop(
        None,
        key=Prop.UPDATE_DATE.value,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat(timespec="milliseconds") if d else None,
        repr=lambda v: v.isoformat(timespec="milliseconds") if v else "None",
    )
    subject: Optional[str] = prop(None)
    requester: Optional["LocalUser"] = prop(None)
    domain: Optional["Domain"] = prop(None)


@propify
class TicketDraftRequestSearch(RequestSearch):
    ticket_draft_id: Optional[int] = prop(None)


@propify
class TicketRequestSearch(RequestSearch):
    from pytos2.securechange.saved_search import DetailedQueryFilters

    ticket_id: Optional[int] = prop(None)
    status: Optional[DetailedQueryFilters.TicketStatus] = prop(
        None, kwargify=str
    )  # Ticketstatus
    current_step: Optional["CurrentStage"] = prop(None)
    application_details: Optional["ApplicationDetails"] = prop(None)
    required_attentions: List["RequiredAttention"] = prop(factory=list)
    completion_ticket_details: Optional["CompletionTicketDetails"] = prop(None)


@propify
class Domain(Jsonable):
    id: Optional[int] = prop(None)
    name: Optional[str] = prop(None)


@propify
class AbstractUserRequest(Jsonable):
    class PartyType(Enum):
        LOCAL_USER = "LOCAL_USER"
        USER = "USER"

    id: Optional[int] = prop(None)
    name: Optional[str] = prop(None)
    type: Optional[PartyType] = prop(None, kwargify=str)  # PartyTypeDTO


@propify
class User(AbstractUserRequest):
    memberships: List["AbstractGroupSearch"] = prop(factory=list)


@propify
class LocalUser(User):
    first_name: Optional[str] = prop(None)
    last_name: Optional[str] = prop(None)
    user: Optional[str] = prop(None)


@propify
class AbstractGroupSearch(Jsonable):
    class PartyType(Enum):
        LOCAL_GROUP = "LOCAL_GROUP"
        GROUP = "GROUP"

    id: Optional[int] = prop(None)
    name: Optional[str] = prop(None)
    type: Optional[PartyType] = prop(None, kwargify=str)  # PartyTypeDTO
    memberships: List["AbstractGroupSearch"] = prop(factory=list)


@propify
class Group(AbstractGroupSearch):
    pass


@propify
class LocalGroup(Group):
    pass


@propify
class CurrentStage(Jsonable):
    id: Optional[int] = prop(None)
    name: Optional[str] = prop(None)
    order: Optional[int] = prop(None)
    tasks: List["TicketTask"] = prop(factory=list)


@propify
class TicketTask(Jsonable):
    id: Optional[int] = prop(None)
    name: Optional[str] = prop(None)
    status: Optional["TicketTaskStatus"] = prop(None)
    handler: Optional["PartySearch"] = prop(None)
    assigned_group: Optional["AbstractGroupSearch"] = prop(None)
    participants: List["PartySearch"] = prop(factory=list)
    unread_comments: Optional[bool] = prop(None)
    task_business_duration: Optional[str] = prop(None)


@propify
class TicketTaskStatus(Jsonable):
    reason: Optional[str] = prop(None)
    status: Optional[str] = prop(None)  # TaskStatusSearchEnumDTO


@propify
class PartySearch(Jsonable):
    class PartyType(Enum):
        LOCAL_USER = "LOCAL_USER"
        USER = "USER"
        LOCAL_GROUP = "LOCAL_GROUP"
        GROUP = "GROUP"

    id: Optional[int] = prop(None)
    name: Optional[str] = prop(None)
    type: Optional[PartyType] = prop(None, kwargify=str)  # PartyTypeDTO
    handler: Optional[LocalUser] = prop(None)
    memberships: List[AbstractGroupSearch] = prop(factory=list)
    first_name: Optional[str] = prop(None)
    last_name: Optional[str] = prop(None)
    user: Optional[str] = prop(None)


@propify
class ApplicationDetails(Jsonable):
    id: Optional[int] = prop(None)
    name: Optional[str] = prop(None)
    domain: Optional["Domain"] = prop(None)


@propify
class RequiredAttention(Jsonable):
    class RequiredAttentionType(Enum):
        MORE_INFORMATION = "MORE_INFORMATION"
        REJECTED = "REJECTED"
        CONFIRM = "CONFIRM"
        RESUBMIT = "RESUBMIT"
        EXPIRED = "EXPIRED"
        ABOUT_TO_EXPIRE = "ABOUT_TO_EXPIRE"

    type: Optional[RequiredAttentionType] = prop(
        None, kwargify=str
    )  # RequiredAttentionType


@propify
class CompletionTicketDetails(Jsonable):
    user: Optional["AbstractUserRequest"] = prop(None)
    last_active_step: Optional["CurrentStage"] = prop(None)



from typing import Optional, Tuple
from .models import GraphResponse, Variables

from ..api import GraphqlError, OAuth2APISession


class BaseGql:
    hostname: Optional[str]
    username: Optional[str]
    password: Optional[str]
    session: Optional[OAuth2APISession]

    def raw_query(self, query: str, variables: Variables = None) -> GraphResponse:
        """
        Execute a GraphQL query and return the raw GraphQL response as a python dictionary.

        Usage:
            # Example 1:
            data = st.graphql.raw_query("query rules($first: Int) { rules { values(first: $first, offset: 0) { id } } }", variables={"first": 100})

        """
        json = (
            {"query": query, "variables": variables} if variables else {"query": query}
        )

        response = self.session.post("graphql", json=json)
        response.raise_for_status()

        res = response.json()
        if "errors" in res:
            raise GraphqlError(errors=res["errors"])

        return response.json()



from enum import Enum
from typing import Optional

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, stringify_optional_obj
from netaddr import IPAddress


@propify
class GenericIgnoredInterface(Jsonable):
    id: Optional[int] = prop(None, converter=optional(int), repr=False)
    interface_name: str = prop(None, key="interfaceName")
    device_id: Optional[int] = prop(None, converter=optional(int), key="mgmtId")
    ip: Optional[IPAddress] = prop(
        None,
        jsonify=stringify_optional_obj,
        converter=optional(IPAddress),
    )



from typing import Optional, List

from enum import Enum

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, kwargify, get_api_node


class UserXsiType(Enum):
    LOCALUSER = "localuser"
    USER = "user"
    GROUP = "group"


class OriginType(Enum):
    LOCAL = "Local"
    LDAP = "LDAP"


@propify
class LDAPConfiguration(Jsonable):
    name: str = prop(None)
    id: int = prop(None)


@propify
class UserDomain(Jsonable):
    name: str = prop()
    id: int = prop()


@propify
class Link(Jsonable):
    class Prop(Enum):
        HREF = "@href"

    href: str = prop(key=Prop.HREF.value)


@propify
class Permission(Jsonable):
    name: Optional[str] = prop(None)
    value: bool = prop(False)
    key: Optional[str] = prop(None, repr=False)


@propify
class UserRole(Jsonable):
    id: int = prop()
    name: str = prop()
    description: Optional[str] = prop(None)
    permissions: List[Permission] = prop(factory=list, flatify="permission", repr=False)
    role_link: Optional[Link] = prop(None, repr=False)


@propify
class UserUpdateRole(Jsonable):
    name: str = prop()


@propify
class SCWPartyLink(Jsonable):
    link: Optional[Link] = prop(None)
    type: Optional[str] = prop(None)
    name: Optional[str] = prop(None)
    id: int = prop(0)
    xsi_type: UserXsiType = prop(UserXsiType.USER, key=Jsonable.Prop.XSI_TYPE.value)


@propify
class SCWParty(Jsonable):
    class Prop(Enum):
        LDAP_DN = "ldapDn"

    class PartyType(Enum):
        USER = "user"
        GROUP = "group"

    class OriginType(Enum):
        LOCAL = "Local"
        LDAP = "LDAP"

    id: int = prop(0)

    xsi_type: UserXsiType = prop(UserXsiType.USER, key=Jsonable.Prop.XSI_TYPE.value)

    email: Optional[str] = prop(None)
    ldap_dn: Optional[str] = prop(None, key=Prop.LDAP_DN.value)
    ldap_configuration: Optional[LDAPConfiguration] = prop(None)
    authentication_method: Optional[str] = prop(None)
    origin_type: Optional[OriginType] = prop(None)
    member_of: List[SCWPartyLink] = prop(factory=list, flatify="user")
    domains: List[UserDomain] = prop(factory=list, flatify="domain")
    name: Optional[str] = prop(None)
    type: Optional[PartyType] = prop(None)
    link: Optional[Link] = prop(None)
    roles: List[UserRole] = prop(factory=list, flatify="role")

    origin_type: Optional[OriginType] = prop(None)


@propify
class SCWUser(SCWParty):
    class Meta(Enum):
        ROOT = "user"
        ENTITY = "user"

    out_of_office_from: Optional[str] = prop(None)
    out_of_office_until: Optional[str] = prop(None)
    notes: Optional[str] = prop(None)
    phone: Optional[str] = prop(None)
    send_email: bool = prop(False)
    first_name: Optional[str] = prop(None)
    last_name: Optional[str] = prop(None)
    display_name: Optional[str] = prop(None)


@propify
class SCWLocalUser(SCWParty):
    email: Optional[str] = prop(None)
    first_name: Optional[str] = prop(None)
    last_name: Optional[str] = prop(None)
    password: str = prop("")
    default_authentication: Optional[bool] = prop(False)
    phone: Optional[str] = prop(None)
    notes: Optional[str] = prop(None)
    name: str = prop("")


@propify
class NewLocalUser(SCWLocalUser):
    origin_type: OriginType = prop(OriginType.LOCAL)


@propify
class GroupPermission(Jsonable):
    name: str = prop("")
    value: bool = prop(False)


@propify
class SCWGroup(SCWParty):
    class Meta(Enum):
        ROOT = "group"
        ENTITY = "group"

    class Prop(Enum):
        GROUP_PERMISSIONS = "groupPermissions"
        GROUP_PERMISSION = "groupPermission"

    group_permissions: List[GroupPermission] = prop(
        factory=list,
        key=Prop.GROUP_PERMISSIONS.value,
        flatify=Prop.GROUP_PERMISSION.value,
    )
    description: Optional[str] = prop(None)
    members: List[SCWPartyLink] = prop(factory=list, flatify="user")


def classify_user_object(obj: dict, obj_type: str = None):
    """
    obj is the dictionary fetched from the server.

    obj_type was added because the `users/{id}` endpoint does not have
    an "@xsi.type" defined on the user object like you would expect.
    """

    if not obj_type:
        obj_type = obj[Jsonable.Prop.XSI_TYPE.value]

    cls = {
        UserXsiType.LOCALUSER.value: SCWUser,
        UserXsiType.USER.value: SCWUser,
        UserXsiType.GROUP.value: SCWGroup,
    }.get(obj_type, SCWParty)

    return cls.kwargify(obj)


@propify
class LDAPParty(Jsonable):
    ldap_dn: str = prop(None)
    ldap_configuration_name: str = prop(None)


@propify
class LDAPSimulateLogin(Jsonable):
    name: str = prop(None)
    origin_type: OriginType = prop(OriginType.LDAP)
    ldap_configuration: Optional[LDAPConfiguration] = prop(None)


@propify
class LDAPUserSimulateLogin(LDAPSimulateLogin):  #
    first_name: Optional[str] = prop(None)
    last_name: Optional[str] = prop(None)
    email: Optional[str] = prop(None)
    phone: Optional[str] = prop(None)
    notes: Optional[str] = prop(None)



import pytest
import json
from pytos2.securechange.fields import (
    classify_service_type,
    AccessRequest,
    MultiAccessRequest,
    Object,
    Service,
    ExistService,
    TCPService,
    UDPService,
    ICMPService,
    OtherService,
    ServiceGroup,
    ExistTCPService,
    ExistUDPService,
    ExistICMPService,
    ExistOtherService,
    PredefinedService,
    PredefinedTCPService,
    PredefinedUDPService,
    PredefinedOtherService,
    ApplicationIdentity,
)
from pytos2.securechange.service import ApplicationIdentityName, PredefinedServiceName


class TestAccessRequest:
    @pytest.fixture
    def access_requests(self):
        j = json.load(
            open(
                "tests/securechange/json/field/multi_access_request_with_all_service_types.json"
            )
        )
        return MultiAccessRequest.kwargify(j)

    def test_ar_inheritance(self, access_requests):
        ar_list = access_requests.access_requests
        assert isinstance(ar_list, list)
        assert isinstance(ar_list[0].services, list)


class TestAccessRequestServices:
    @pytest.fixture
    def ar_services(self):
        j = json.load(
            open(
                "tests/securechange/json/field/multi_access_request_with_all_service_types.json"
            )
        )
        return MultiAccessRequest.kwargify(j).access_requests[0].services

    def test_service_inheritance(self, ar_services):
        assert isinstance(ar_services, list)
        assert isinstance(ar_services[0], TCPService)
        assert isinstance(ar_services[0], Service)

        assert isinstance(ar_services[1], UDPService)
        assert isinstance(ar_services[1], Service)

        assert isinstance(ar_services[2], ICMPService)
        assert isinstance(ar_services[2], Service)

        assert isinstance(ar_services[3], ExistICMPService)
        assert isinstance(ar_services[3], ExistService)
        assert isinstance(ar_services[3], Object)
        assert isinstance(ar_services[3], ICMPService)
        assert isinstance(ar_services[3], Service)

        assert isinstance(ar_services[4], ApplicationIdentity)
        assert isinstance(ar_services[4], Service)
        assert isinstance(ar_services[4].services, list)
        assert isinstance(ar_services[4].services[0], TCPService)
        assert isinstance(ar_services[5], ServiceGroup)
        assert isinstance(ar_services[5], ExistService)
        assert isinstance(ar_services[5], Object)
        assert isinstance(ar_services[5].members, list)

        assert isinstance(ar_services[7], ExistOtherService)
        assert isinstance(ar_services[7], ExistService)
        assert isinstance(ar_services[7], Object)
        assert isinstance(ar_services[7], OtherService)
        assert isinstance(ar_services[7], Service)

        assert isinstance(ar_services[8], ExistOtherService)
        assert isinstance(ar_services[8], ExistService)
        assert isinstance(ar_services[8], Object)
        assert isinstance(ar_services[8], OtherService)
        assert isinstance(ar_services[8], Service)

        assert isinstance(ar_services[9], ExistUDPService)
        assert isinstance(ar_services[9], ExistService)
        assert isinstance(ar_services[9], Object)
        assert isinstance(ar_services[9], UDPService)
        assert isinstance(ar_services[9], Service)

        assert isinstance(ar_services[10], ExistTCPService)
        assert isinstance(ar_services[10], ExistService)
        assert isinstance(ar_services[10], Object)
        assert isinstance(ar_services[10], TCPService)
        assert isinstance(ar_services[10], Service)

        assert isinstance(ar_services[14], PredefinedTCPService)
        assert isinstance(ar_services[14], PredefinedService)
        assert isinstance(ar_services[14], TCPService)
        assert isinstance(ar_services[14], Service)

        assert isinstance(ar_services[15], PredefinedOtherService)
        assert isinstance(ar_services[15], PredefinedService)
        assert isinstance(ar_services[15], OtherService)
        assert isinstance(ar_services[15], Service)

    def test_service_attributes(self, ar_services):
        assert ar_services[0].protocol == "TCP"
        assert ar_services[0].port == "111"
        assert ar_services[0].at_type.value == "PROTOCOL"
        assert ar_services[0].id == 22157

        assert ar_services[1].protocol == "UDP"
        assert ar_services[1].port == "222"
        assert ar_services[1].at_type.value == "PROTOCOL"
        assert ar_services[1].id == 22158

        assert ar_services[2].protocol == "ICMP"
        assert ar_services[2].type == "253"
        assert ar_services[2].at_type.value == "PROTOCOL"
        assert ar_services[2].id == 22870

        assert ar_services[3].at_type.value == "Object"
        assert ar_services[3].id == 22159
        assert ar_services[3].name == "administratively-prohibited"
        assert ar_services[3].management_id == 7
        assert ar_services[3].management_name == "RTR2"
        assert ar_services[3].uid == "{c8944d28-dbef-44ef-a49f-e33a4c09ebca}"
        assert ar_services[3].protocol == "icmp"
        assert ar_services[3].type == "3"
        assert ar_services[3].obj_type == "icmp"

        assert ar_services[4].at_type.value == "APPLICATION_IDENTITY"
        assert ar_services[4].id == 22160
        assert ar_services[4].name == ApplicationIdentityName._BAO
        assert hasattr(ar_services[4], "services")
        assert ar_services[4].services[0].protocol == "TCP"
        assert ar_services[4].services[0].port == "3468"
        assert ar_services[4].services[0].at_type.value == "PROTOCOL"
        assert ar_services[4].services[1].protocol == "TCP"
        assert ar_services[4].services[1].port == "11300"
        assert ar_services[4].services[1].at_type.value == "PROTOCOL"

        assert ar_services[5].at_type.value == "Object"
        assert ar_services[5].id == 22161
        assert ar_services[5].name == "daytime"
        assert ar_services[5].management_id == 243
        assert ar_services[5].management_name == "CMA-R80"
        assert ar_services[5].uid == "{97AEB46F-9AEA-11D5-BD16-0090272CCB30}"
        assert ar_services[5].members[0] == "daytime-tcp"
        assert ar_services[5].members[1] == "daytime-udp"
        assert ar_services[5].obj_type == "group"

        assert ar_services[6].at_type.value == "Object"
        assert ar_services[6].id == 22162
        assert ar_services[6].name == "Data Recovery Appliance"
        assert ar_services[6].management_id == 59
        assert ar_services[6].management_name == "NSX-Distributed Firewall"
        assert ar_services[6].uid == "{32efb97a-bb5c-4bc8-8161-3b1e1846d093}"
        assert ar_services[6].members[0] == "VMware-DataRecovery"
        assert ar_services[6].members[1] == "VMware-ESXi5.x-TCP"
        assert ar_services[6].members[2] == "HTTPS"
        assert ar_services[6].obj_type == "group"

        assert ar_services[7].at_type.value == "Object"
        assert ar_services[7].id == 22163
        assert ar_services[7].name == "ah"
        assert ar_services[7].management_id == 32
        assert ar_services[7].management_name == "RTR4"
        assert ar_services[7].uid == "{3fbd8116-3801-4bee-8593-3cbf999da671}"
        assert ar_services[7].obj_type == "other"
        assert ar_services[7].type == "51"

        assert ar_services[9].at_type.value == "Object"
        assert ar_services[9].id == 22585
        assert ar_services[9].name == "archie"
        assert ar_services[9].management_id == 20
        assert ar_services[9].management_name == "CP SMC"
        assert ar_services[9].uid == "{97AEB3D6-9AEA-11D5-BD16-0090272CCB30}"
        assert ar_services[9].protocol == "udp"
        assert ar_services[9].port == "1525"
        assert ar_services[9].obj_type == "udp"

        assert ar_services[10].at_type.value == "Object"
        assert ar_services[10].id == 22584
        assert ar_services[10].name == "AP-Defender"
        assert ar_services[10].management_id == 20
        assert ar_services[10].management_name == "CP SMC"
        assert ar_services[10].uid == "{97AEB3E9-9AEA-11D5-BD16-0090272CCB30}"
        assert ar_services[10].protocol == "tcp"
        assert ar_services[10].port == "2626"
        assert ar_services[10].obj_type == "tcp"

        assert ar_services[14].protocol == "TCP"
        assert ar_services[14].port == "443"
        assert ar_services[14].at_type.value == "PREDEFINED"
        assert ar_services[14].id == 22168

        assert ar_services[15].type == "51"
        assert ar_services[15].protocol == "OTHER"
        assert ar_services[15].name == PredefinedServiceName.AH
        assert ar_services[15].at_type.value == "PREDEFINED"
        assert ar_services[15].id == 22196



from pytos2.utils import propify, prop
from pytos2.models import Jsonable

from netaddr import IPAddress, IPRange

from typing import List, Optional
from enum import Enum

from attr.converters import optional


def _classify(obj, type_dict, default_type):
    xsi_type = obj.get(Jsonable.Prop.XSI_TYPE.value)

    cls = type_dict.get(xsi_type, None)
    if cls:
        return cls.kwargify(obj)
    else:
        return default_type.kwargify(obj)


def classify_violation(obj):
    type_dict = {
        SecurityPolicyViolationType.SECURITY_ZONE_MATRIX_VIOLATION.value: SecurityZoneMatrixViolation,
        SecurityPolicyViolationType.UNMATCHED_PRIVATE_ADDRESS_VIOLATION.value: UnmatchedPrivateAddressViolation,
    }

    return _classify(obj, type_dict, SecurityPolicyViolation)


def classify_violation_network_object(obj):
    type_dict = {
        NetworkObjectType.IP_NETWORK_OBJECT.value: IpNetworkObject,
        NetworkObjectType.RANGE_NETWORK_OBJECT.value: RangeNetworkObject,
        NetworkObjectType.ANY_NETWORK_OBJECT.value: AnyNetworkObject,
        NetworkObjectType.SINGLE_NETWORK_OBJECT.value: SingleNetworkObject,
        NetworkObjectType.GROUP_MEMBER_NETWORK_OBJECT.value: GroupMemberNetworkObject,
        NetworkObjectType.LDAP_ENTITY.value: LDAPEntity,
    }

    return _classify(obj, type_dict, NetworkObjectInViolation)


def classify_matrix_cell(obj):
    type_dict = {
        MatrixCellType.BLOCKED_CELL_VIOLATION.value: BlockedCellViolation,
        MatrixCellType.RESTRICTED_CELL.value: RestrictedCellViolation,
    }

    return _classify(obj, type_dict, MatrixCellViolation)


def classify_violation_service(obj):
    type_dict = {
        ServiceType.ANY_SERVICE.value: AnyService,
        ServiceType.SINGLE_SERVICE.value: None,
        ServiceType.SINGLE_SERVICE_OBJECT.value: None,
        ServiceType.GROUP_MEMBER_SERVICE_OBJECT.value: None,
    }

    return _classify(obj, type_dict, ServiceInViolation)


@propify
class SecurityZoneMatrix(Jsonable):
    name: str = prop()


class SecurityPolicyViolationType(Enum):
    SECURITY_ZONE_MATRIX_VIOLATION = "security_zone_matrix_violation"
    UNMATCHED_PRIVATE_ADDRESS_VIOLATION = "unmatched_private_address_violation"


class ServiceType(Enum):
    ANY_SERVICE = "any_service"
    SINGLE_SERVICE = "single_service"
    SINGLE_SERVICE_OBJECT = "single_service_object"
    GROUP_MEMBER_SERVICE_OBJECT = "group_member_service_object"


class NetworkObjectType(Enum):
    IP_NETWORK_OBJECT = "ip_network_object"
    RANGE_NETWORK_OBJECT = "range_network_object"

    ANY_NETWORK_OBJECT = "any_network_object"
    SINGLE_NETWORK_OBJECT = "single_network_object"
    GROUP_MEMBER_NETWORK_OBJECT = "group_member_network_object"
    LDAP_ENTITY = "ldap_entity"


class MatrixCellType(Enum):
    BLOCKED_CELL_VIOLATION = "blocked_cell_violation"
    RESTRICTED_CELL = "restricted_cell"


@propify
class ServiceInViolation(Jsonable):
    name: Optional[str] = prop(None)


@propify
class AnyService(ServiceInViolation):
    pass


@propify
class SingleService(ServiceInViolation):
    protocol: Optional[str] = prop(None)
    port: Optional[int] = prop(None, converter=optional(int))


@propify
class SingleServiceObject(ServiceInViolation):
    uid: Optional[str] = prop(None)
    management_id: Optional[int] = prop(None)


@propify
class GroupMemberServiceObject(ServiceInViolation):
    group_member: Optional[SingleServiceObject] = prop(None)
    group_member_path: Optional[str] = prop(None)


@propify
class NetworkObjectInViolation(Jsonable):
    name: Optional[str] = prop(None)


class AnyNetworkObject(NetworkObjectInViolation):
    pass


@propify
class SingleNetworkObject(NetworkObjectInViolation):
    uid: Optional[str] = prop(None)
    management_id: Optional[int] = prop(None)


@propify
class GroupMemberNetworkObject(NetworkObjectInViolation):
    group_member: Optional[SingleNetworkObject] = prop(None)
    group_member_path: Optional[str] = prop(None)
    name: Optional[str] = prop(None)


@propify
class LDAPEntity(NetworkObjectInViolation):
    ldap_dn: Optional[str] = prop(None)
    ldap_id: Optional[str] = prop(None)
    name: Optional[str] = prop(None)


@propify
class IpNetworkObject(NetworkObjectInViolation):
    ip: Optional[IPAddress] = prop(None)
    mask: Optional[str] = prop(None)
    name: Optional[str] = prop(None)


@propify
class RangeNetworkObject(NetworkObjectInViolation):
    min_ip: Optional[IPAddress] = prop(None)
    max_ip: Optional[IPAddress] = prop(None)

    @property
    def ip_range(self) -> Optional[IPRange]:
        if self.min_ip and self.max_ip:
            return IPRange(self.min_ip, self.max_ip)


@propify
class MatrixCellViolation(Jsonable):
    from_zone: Optional[str] = prop(None)
    to_zone: Optional[str] = prop(None)

    sources: List[NetworkObjectInViolation] = prop(
        factory=list, flatify="source", kwargify=classify_violation_network_object
    )
    destinations: List[NetworkObjectInViolation] = prop(
        factory=list, flatify="destination", kwargify=classify_violation_network_object
    )

    allowed_services: List[ServiceInViolation] = prop(
        factory=list, flatify="allowed_service", kwargify=classify_violation_service
    )
    not_allowed_services: List[ServiceInViolation] = prop(
        factory=list, flatify="not_allowed_service", kwargify=classify_violation_service
    )
    blocked_services: List[ServiceInViolation] = prop(
        factory=list, flatify="blocked_service", kwargify=classify_violation_service
    )
    not_blocked_services: List[ServiceInViolation] = prop(
        factory=list, flatify="not_blocked_service", kwargify=classify_violation_service
    )


@propify
class BlockedCellViolation(MatrixCellViolation):
    pass


class RestrictedCellViolation(MatrixCellViolation):
    class Prop(Enum):
        VIOLATING_SERVICES = "violatingServices"
        REQUIREMENT_SERVICES = "requirementServices"

    violating_services: List[ServiceInViolation] = prop(
        factory=list, key=Prop.VIOLATING_SERVICES.value
    )
    flow: Optional[str] = prop(None)
    requirement_services: List[ServiceInViolation] = prop(
        factory=list, key=Prop.REQUIREMENT_SERVICES.value
    )
    flow_sources: List[NetworkObjectInViolation] = prop(
        factory=list, flatify="flow_source"
    )
    flow_destination: List[NetworkObjectInViolation] = prop(
        factory=list, flatify="flow_destination"
    )


@propify
class SecurityPolicyViolation(Jsonable):
    class Severity(Enum):
        CRITICAL = "CRITICAL"
        HIGH = "HIGH"
        MEDIUM = "MEDIUM"
        LOW = "LOW"

    severity: Severity = prop()


@propify
class SecurityZoneMatrixViolation(SecurityPolicyViolation):
    security_zone_matrix: Optional[SecurityZoneMatrix] = prop(None)
    matrix_cell_violation: Optional[MatrixCellViolation] = prop(
        None, kwargify=classify_matrix_cell
    )


@propify
class UnmatchedPrivateAddressViolation(SecurityPolicyViolation):
    violation_source_networks: List[NetworkObjectInViolation] = prop(
        factory=list,
        kwargify=classify_violation_network_object,
        flatify="violation_source_network",
    )
    violation_destination_networks: List[NetworkObjectInViolation] = prop(
        factory=list,
        kwargify=classify_violation_network_object,
        flatify="violation_destination_network",
    )


@propify
class ComplianceRule(Jsonable):
    name: Optional[str] = prop(None)
    number: Optional[str] = prop(None)


@propify
class CompliancePolicy(Jsonable):
    compliance_rules: List[ComplianceRule] = prop(
        factory=list, flatify="compliance_rule"
    )
    name: Optional[str] = prop(None)
    type: Optional[str] = prop(None)


@propify
class RiskAnalysisResult(Jsonable):
    class RiskStatus(Enum):
        HAS_RISK = "has risk"
        NO_RISK = "no risk"
        NOT_RUN = "not run"

    status: RiskStatus = prop()
    compliance_policies: List[CompliancePolicy] = prop(
        factory=list, flatify="compliance_policy"
    )
    security_policy_violations: List[SecurityPolicyViolation] = prop(
        factory=list, flatify="security_policy_violation", kwargify=classify_violation
    )



from datetime import datetime
from typing import Optional, List

from pytos2.utils import propify, prop, safe_iso8601_date, safe_date, TimeFormat
from pytos2.models import Jsonable


@propify
class ChangeWindow(Jsonable):
    # Rekeying uuid to Jsonable.id, since that's our standard ID prop.
    id: str = prop("", key="uuid")
    name: str = prop("")
    description: str = prop("")
    domain_id: int = prop()
    enabled: bool = prop(False)

    @property
    def uuid(self):
        return self.id

    @uuid.setter
    def uuid(self, v):
        self.id = v


@propify
class DeviceCommitDevice(Jsonable):
    status: str = prop("")
    revision_id: int = prop()
    warnings: List[str] = prop(factory=list, flatify="warning")


@propify
class DeviceCommitResult(Jsonable):
    errors: List[str] = prop(factory=list, flatify="error")
    device: DeviceCommitDevice = prop()


@propify
class DeviceCommit(Jsonable):
    result: DeviceCommitResult = prop()


@propify
class ChangeWindowTask(Jsonable):
    start_date: Optional[datetime] = prop(
        None,
        repr=False,
        jsonify=False,
        kwargify=lambda val: safe_date(val, TimeFormat.UTC),
    )
    end_date: Optional[datetime] = prop(
        None,
        repr=False,
        jsonify=False,
        kwargify=lambda val: safe_date(val, TimeFormat.UTC),
    )
    errors: List[str] = prop(factory=list, flatify="error")
    device_commits: List[DeviceCommit] = prop(factory=list, flatify="device_commit")



import pytest
from netaddr import IPNetwork, IPRange, IPAddress
from . import conftest  # noqa
from pytos2.securechange.fields import MultiAccessRequest


@pytest.fixture
def ars(get_test_field):
    return get_test_field(MultiAccessRequest).access_requests


class TestAccessRequest:
    @pytest.mark.parametrize(
        "attr", ["targets", "users", "sources", "destinations", "services", "labels"]
    )
    def test_list_attrs(self, ars, attr):
        assert isinstance(getattr(ars[0], attr), list)


class TestAccessRequestIP:
    @pytest.fixture
    def subnet(self, ars):
        return ars[1].sources[3]

    def test_subnet(self, subnet):
        assert str(subnet.subnet) == "10.20.50.0/24"

    def test_set_subnet_obj(self, subnet):
        subnet.subnet = IPNetwork("1.2.3.4/31")
        assert subnet.ip_address == "1.2.3.4"
        assert subnet.netmask == "255.255.255.254"
        assert subnet.cidr == 31

    def test_set_subnet_str(self, subnet):
        subnet.subnet = "1.2.3.0/30"
        assert subnet.netmask == "255.255.255.252"
        subnet.subnet = "1.2.3.0/255.255.255.0"
        assert subnet.cidr == 24

    def test_set_invalid_subnet(self, subnet):
        with pytest.raises(ValueError):
            subnet.subnet = "thing"


class TestAccessRequestNatIP:
    @pytest.fixture
    def nat_ip(self, ars):
        return ars[1].sources[4]

    def test_subnet(self, nat_ip):
        assert str(nat_ip.subnet) == "5.6.7.8/32"

    def test_set_subnet_obj(self, nat_ip):
        nat_ip.subnet = IPNetwork("1.2.3.4/31")
        assert nat_ip.ip_address == "1.2.3.4"
        assert nat_ip.netmask == "255.255.255.254"
        assert nat_ip.cidr == 31

    def test_set_nat_subnet_obj(self, nat_ip):
        nat_ip.nat_subnet = IPNetwork("1.2.3.4/31")
        assert nat_ip.nat_ip_address == "1.2.3.4"
        assert nat_ip.nat_netmask == "255.255.255.254"
        assert nat_ip.nat_cidr == 31

    def test_set_subnet_str(self, nat_ip):
        nat_ip.subnet = "1.2.3.0/30"
        assert nat_ip.netmask == "255.255.255.252"
        nat_ip.subnet = "1.2.3.0/255.255.255.0"
        assert nat_ip.cidr == 24

    def test_set_nat_subnet_str(self, nat_ip):
        nat_ip.nat_subnet = "1.2.3.0/30"
        assert nat_ip.nat_ip_address == "1.2.3.0"
        assert nat_ip.nat_netmask == "255.255.255.252"
        assert nat_ip.nat_cidr == 30


class TestAccessRequestRange:
    @pytest.fixture
    def ip_range(self, ars):
        return ars[1].sources[2]

    def test_range(self, ip_range):
        assert str(ip_range.range) == "10.10.10.10-11.1.1.1"

    def test_set_range_obj(self, ip_range):
        ip_range.range = IPRange("1.2.3.4", "1.2.3.6")
        assert ip_range.range_first_ip == "1.2.3.4"
        assert ip_range.range_last_ip == "1.2.3.6"

    def test_set_range_str(self, ip_range):
        ip_range.range = "1.1.1.1-1.2.3.4"
        assert ip_range.range_first_ip == "1.1.1.1"
        assert ip_range.range_last_ip == "1.2.3.4"

    def test_set_range_iter(self, ip_range):
        ip_range.range = ("1.0.0.0", "1.4.3.2")
        assert ip_range.range_first_ip == "1.0.0.0"
        assert ip_range.range_last_ip == "1.4.3.2"

    def test_set_invalid_range(self, ip_range):
        with pytest.raises(ValueError):
            ip_range.range = "thing"


class TestAccessRequestDNS:
    @pytest.fixture
    def dns(self, ars):
        return ars[1].sources[1]

    def test_dns_ip_addresses(self, dns):
        assert all(isinstance(ip, IPAddress) for ip in dns.dns_ip_addresses)

    def test_json_override(self, dns):
        dns._json = {}
        assert dns._json == {}



import pytos2
from .api import ScwAPI
from .entrypoint import Scw
from .ticket import Ticket
from .trigger import Trigger
from .ticket_requests import RequestsSearchList


__all__ = ["ScwAPI", "Ticket", "Trigger", "Scw", "RequestsSearchList"]



import os
import warnings

SHOW_DEPRECATION_WARNINGS = os.getenv("SHOW_DEPRECATION_WARNINGS")

if not bool(SHOW_DEPRECATION_WARNINGS):
    warnings.filterwarnings("ignore", category=DeprecationWarning)



import pytest
import responses

from pytos2.securetrack.properties import Properties


class TestProperties:
    @responses.activate
    def test_get_properties(self, properties_mock, st):
        properties = st.get_properties()

        assert properties.secure_change_addresses[0].ip_address == "127.0.0.1"
        assert properties.secure_change_addresses[0].type == "external"
        assert properties.secure_change_addresses[1].ip_address == "127.0.0.1"
        assert properties.secure_change_addresses[1].type == "internal"

        assert (
            properties.general_properties[0].key == "LICENSE_ABOUT_TO_EXPIRE_THRESHOLD"
        )
        assert properties.general_properties[0].value == "45"

    @responses.activate
    def test_set_license_notification_days(self, properties_mock, st):
        st.set_license_notification_days(100)



from enum import Enum
from datetime import date, datetime
from typing import Optional, List

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, safe_iso8601_date


@propify
class Customer(Jsonable):
    id: str = prop("", repr=False)
    site_id: Optional[str] = prop(None, repr=False)
    site: Optional[str] = prop(None)
    name: str = prop("")


@propify
class SKUDevice(Jsonable):
    id: str = prop("")
    consumed: int = prop(0)
    name: str = prop("")


@propify
class SKU(Jsonable):
    name: str = prop("")
    description: Optional[str] = prop(None, repr=False)
    quantity: int = prop(0)
    expiration: Optional[datetime] = prop(
        None, kwargify=safe_iso8601_date, jsonify=lambda d: d.isoformat() if d else None
    )
    devices: List[SKUDevice] = prop(factory=list, key="devices", repr=False)


@propify
class License(Jsonable):
    id: str = prop("")
    expiration: datetime = prop(
        None, kwargify=safe_iso8601_date, jsonify=lambda d: d.isoformat() if d else None
    )
    uid: Optional[str] = prop(None, repr=False)
    customer: Customer = prop(None)
    issued: Optional[datetime] = prop(
        None, kwargify=safe_iso8601_date, jsonify=lambda d: d.isoformat() if d else None
    )
    skus: List[SKU] = prop(factory=list, repr=False, flatify="sku")
    type: str = prop("")


@propify
class TieredLicenseMessage(Jsonable):
    category: str = prop(None)
    severity: str = prop(None)
    code: str = prop(None)
    message: str = prop(None)
    params: List[str]


@propify
class TieredLicense(Jsonable):
    class LicenseType(Enum):
        EVALUATION = "evaluation"
        FULL = "full"

    class Status(Enum):
        VALID = "valid"
        EXPIRED = "expired"

    class SiteType(Enum):
        PRODUCTION = "production"
        LAB = "lab"

    supported: bool = prop(None)
    version: str = prop(None)
    tier: str = prop(None)
    customer_id: str = prop(None)
    customer_name: str = prop(None)
    license_file_id: str = prop(None)
    site_name: str = prop(None)
    site_type: SiteType = prop(None)
    site_id: str = prop(None)
    installed: datetime = prop(
        None, kwargify=safe_iso8601_date, jsonify=lambda d: d.isoformat() if d else None
    )
    expires: datetime = prop(
        None, kwargify=safe_iso8601_date, jsonify=lambda d: d.isoformat() if d else None
    )
    status: Status = prop(None)
    type: LicenseType = prop(None)
    messages: List[TieredLicenseMessage] = prop(list)



from enum import Enum
from typing import Optional, List
from datetime import datetime, date, time
from time import mktime

from attr.converters import optional

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, kwargify, TimeFormat

from pytos2 import securetrack


@propify
class RevisionTicket(Jsonable):
    class Source(Enum):
        SCW = "scw"

    # id is handled by Jsonable
    source: Optional[Source] = prop(None)


@propify
class RevisionModuleAndPolicy(Jsonable):
    module: Optional[str] = prop(None)
    policy: Optional[str] = prop(None)


@propify
class Revision(Jsonable):
    class Action(Enum):  # tee-hee class action.
        AUTOMATIC = "automatic"
        INSTALLED = "installed"
        SAVED = "saved"

    class Prop(Enum):
        REVISION_ID = "revisionId"
        DATE = "date"
        TIME = "time"
        GUI_CLIENT = "guiClient"
        AUDIT_LOG = "auditLog"
        POLICY_PACKAGE = "policyPackage"
        AUTHORIZATION_STATUS = "authorizationStatus"
        READY = "ready"
        MODULES_AND_POLICY = "modules_and_policy"
        TICKETS = "tickets"

    class AuthorizationStatus(Enum):
        AUTHORIZED = "authorized"
        ERROR = "error"
        N_A = "n_a"
        PENDING = "pending"
        UNAUTHORIZED = "unauthorized"

    revision_id: Optional[int] = prop(
        None, key=Prop.REVISION_ID.value, converter=optional(int)
    )
    action: Optional[Action] = prop(None, repr=False)

    _date: Optional[date] = prop(None, repr=False, jsonify="date_str")
    _time: Optional[time] = prop(None, repr=False, jsonify="time_str")

    admin: Optional[str] = prop(None, repr=False)
    gui_client: Optional[str] = prop(None, repr=False, key=Prop.GUI_CLIENT.value)
    audit_log: Optional[str] = prop(None, repr=False, key=Prop.AUDIT_LOG.value)
    policy_package: Optional[str] = prop(
        None, repr=False, key=Prop.POLICY_PACKAGE.value
    )
    authorization_status: Optional[AuthorizationStatus] = prop(
        None, repr=False, key=Prop.AUTHORIZATION_STATUS.value
    )
    modules_and_policy: List[RevisionModuleAndPolicy] = prop(
        factory=list, flatify="module_and_policy", repr=False
    )

    tickets: List[RevisionTicket] = prop(factory=list, flatify="ticket", repr=False)

    firewall_status: bool = prop(False, repr=False)
    is_ready: bool = prop(False, repr=False, key=Prop.READY.value)

    @property
    def date(self):
        return self._date

    @date.setter
    def date(self, val):
        if type(val) is date:
            self._date = val
        elif type(val) is datetime:
            self._date = val.date()
        elif isinstance(val, str):
            self._date = datetime.strptime(val, TimeFormat.DATE.value).date()
        elif val is None:
            self._date = None
        else:
            raise ValueError("Cannot cast date from instance of '{}'".format(type(val)))

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, val):
        if type(val) is time:
            self._time = val
        elif type(val) is datetime:
            self._time = val.time()
        elif isinstance(val, str):
            self._time = datetime.strptime(val, TimeFormat.TIME.value).time()
        elif val is None:
            self._time = None
        else:
            raise ValueError("Cannot cast time from instance of '{}'".format(type(val)))

    @property
    def date_str(self):
        return self._date.strftime(TimeFormat.DATE.value)

    @property
    def time_str(self):
        return self._time.strftime(TimeFormat.TIME.value)[
            :-3
        ]  # Drop the last 3 digits of microseconds to make milliseconds

    @property
    def unix_timestamp(self):
        _timestamp = mktime(self._date.timetuple())
        _timestamp += (
            self._time.hour * 3600 + self._time.minute * 60 + self._time.second
        )
        return _timestamp

    @unix_timestamp.setter
    def unix_timestamp(self, val):
        if not isinstance(val, (float, int)):
            raise ValueError(
                "Cannot cast unix timestamp from instance of '{}'".format(type(val))
            )

        _datetime = datetime.fromtimestamp(val)
        self._date = _datetime.date()
        self._time = _datetime.time()

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)

        _date = _obj.get("date")
        _time = _obj.get("time")

        _date = (
            datetime.strptime(_date, TimeFormat.DATE.value).date()
            if _date is not None
            else None
        )
        _time = (
            datetime.strptime(_time, TimeFormat.TIME.value).time()
            if _time is not None
            else None
        )

        kwargs["date"] = _date
        kwargs["time"] = _time

        return cls(**kwargs)

    def get_rules(self):
        return securetrack.St.default.get_rules(revision=self.id)



from datetime import date, datetime
from typing import Union, Optional, Iterable, Iterator, List
from enum import Enum

from pytos2.models import DateFilterType, coerce_date_filter_type
from traversify import Traverser
from pathlib import Path

# avoid circular imports
import pytos2
from .api import ScwAPI
from requests import Response
from requests.exceptions import (
    ConnectionError,
    HTTPError,
    RequestException,
    Timeout,
    JSONDecodeError,
)
from pytos2.securechange.workflow import BasicWorkflowInfo, WorkflowType, FullWorkflow
from pytos2.securechange.user import (
    classify_user_object,
    SCWParty,
    SCWUser,
    SCWGroup,
    UserRole,
    Permission,
    LDAPParty,
    LDAPSimulateLogin,
    LDAPUserSimulateLogin,
    NewLocalUser,
    UserUpdateRole,
    OriginType,
)

from pytos2.securechange.device import DeviceExclusions
from pytos2.securechange.workflow_triggers import WorkflowTrigger
from pytos2.securechange.extension import MarketplaceApp
from pytos2.securechange.saved_search import Query, FreeTextQuery, DetailedQuery
from pytos2.securechange.domain import Domain
from pytos2.securechange.ticket_requests import (
    RequestsSearchList,
    RequestSearch,
    TicketRequestSearch,
    TicketDraftRequestSearch,
)

from pytos2.utils import NoInstance, get_api_node
from pytos2.utils.cache import Cache, CacheIndex
from pytos2.api import Pager


class Scw:
    default: Union["Scw", NoInstance] = NoInstance(
        "Scw.default",
        "No Scw instance has been initialized yet, initialize with `Scw(*args, **kwargs)`",
    )

    def __init__(
        self,
        hostname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        default=True,
    ):
        self.api: ScwAPI = ScwAPI(hostname, username, password)
        if default:
            Scw.default = self

        self.user_cache = Cache()
        self.users_by_name = self.user_cache.make_index("name")
        self.users_by_id = self.user_cache.make_index("id")

    def ticket_search(
        self,
        subject: Optional[str] = None,
        requester: Optional[str] = None,
        group: Optional[str] = None,
        assigned_to: Optional[str] = None,
        priority: Optional[str] = None,
        status: Optional[Union[str, "pytos2.securechange.ticket.TicketStatus"]] = None,
        sla_status: Optional[
            Union[str, "pytos2.securechange.ticket.Ticket.SlaStatus"]
        ] = None,
        field_name: Optional[str] = None,
        field_value: Optional[str] = None,
        current_step: Optional[str] = None,
        expiration_date_from: Optional[str] = None,
        expiration_date_to: Optional[str] = None,
        domain_name: Optional[str] = None,
    ) -> List["pytos2.securechange.ticket.TicketSearchResult"]:
        """
        Retrieve a list of TicketSearchResult objects for the given optional search arguments

        Method: GET
        URL: /securechangeworkflow/api/securechange/tickets/search
        Version: R22-2+

        Usage:
            # Example #1:
            search_results = scw.ticket_search(subject='"Auto Recert Ticket"')

            # Example #2:
            search_results = scw.ticket_search(requester='ateam', assigned_to='api_user')
        """
        from pytos2.securechange.ticket import TicketSearchResult

        # It's clearer to write out the params explicitly than to use locals()
        params = {
            "subject": subject,
            "requester": requester,
            "group": group,
            "assigned_to": assigned_to,
            "priority": priority,
            "status": status,
            "sla_status": sla_status,
            "field_name": field_name,
            "field_value": field_value,
            "current_step": current_step,
            "expiration_date_from": expiration_date_from,
            "expiration_date_to": expiration_date_to,
            "domain_name": domain_name,
        }

        for k, param in params.items():
            if isinstance(param, Enum):
                params[k] = param.value

        r = self.api.session.get("tickets/search", params=params)
        if not r.ok:
            r.raise_for_status()
        else:
            tickets = get_api_node(
                r.json(), "tickets_search_results.ticket_result", listify=True
            )
            return [TicketSearchResult.kwargify(t) for t in tickets]

    def ticket_search_by_details(
        self,
        subject: Optional[str] = None,
        requester: Optional[str] = None,
        group: Optional[str] = None,
        assigned_to: Optional[str] = None,
        priority: Optional[str] = None,
        task_status: Optional[
            Union[str, "pytos2.securechange.ticket.Task.Status"]
        ] = None,
        status: Optional[Union[str, "pytos2.securechange.ticket.TicketStatus"]] = None,
        sla_status: Optional[
            Union[str, "pytos2.securechange.ticket.Ticket.SlaStatus"]
        ] = None,
        sla_outcome: Optional[
            Union[str, "pytos2.securechange.ticket.SlaOutcome"]
        ] = None,
        field_name: Optional[str] = None,
        field_value: Optional[str] = None,
        current_step: Optional[str] = None,
        expiration_date_from: Optional[str] = None,
        expiration_date_to: Optional[str] = None,
        domain_name: Optional[str] = None,
    ) -> List["pytos2.securechange.ticket.TicketSearchResult"]:
        """
        Retrieve a list of TicketSearchResult objects for the given optional search arguments

        Method: GET
        URL: /securechangeworkflow/api/securechange/tickets/search_by/details
        Version: R23-1+

        Usage:
            # Example #1:
            search_results = scw.ticket_search_by_details(subject='"Auto Recert Ticket"')

            # Example #2:
            search_results = scw.ticket_search_by_details(requester='ateam', assigned_to='api_user')
        """

        from pytos2.securechange.ticket import TicketSearchResult

        # It's clearer to write out the params explicitly than to use locals()
        params = {
            "subject": subject,
            "requester": requester,
            "group": group,
            "assigned_to": assigned_to,
            "priority": priority,
            "task_status": task_status,
            "status": status,
            "sla_status": sla_status,
            "sla_outcome": sla_outcome,
            "field_name": field_name,
            "field_value": field_value,
            "current_step": current_step,
            "expiration_date_from": expiration_date_from,
            "expiration_date_to": expiration_date_to,
            "domain_name": domain_name,
        }

        for k, param in params.items():
            if isinstance(param, Enum):
                params[k] = param.value

        r = self.api.session.get("tickets/search_by/details", params=params)
        r_json = self.api.handle_json(r, "ticket_search_by_details", action="get")
        tickets = get_api_node(
            r_json, "tickets_search_results.ticket_result", listify=True
        )

        return [TicketSearchResult.kwargify(t) for t in tickets]

    def ticket_search_by_saved_search(
        self, saved_search_id: Union[int, Query]
    ) -> List["pytos2.securechange.ticket.TicketSearchResult"]:
        """
        Retrieve a list of TicketSearchResult objects for the given optional search arguments

        Method: GET
        URL: /securechangeworkflow/api/securechange/tickets/search_by/query
        Version: R23-1+

        Usage:
            # Example #1:
            search_results = scw.ticket_search_by_saved_search(saved_search_id=15)
        """
        from pytos2.securechange.ticket import TicketSearchResult

        if isinstance(saved_search_id, Query):
            saved_search_id = saved_search_id.id
        if not isinstance(saved_search_id, int):
            raise TypeError("saved_search_id must be an int")

        params = {
            "query_id": saved_search_id,
        }

        r = self.api.session.get("tickets/search_by/query", params=params)
        r_json = self.api.handle_json(r, "ticket_search_by_query", action="get")
        tickets = get_api_node(
            r_json, "tickets_search_results.ticket_result", listify=True
        )

        return [TicketSearchResult.kwargify(t) for t in tickets]

    def ticket_search_by_group(
        self, group_id: int
    ) -> List["pytos2.securechange.ticket.TicketSearchResult"]:
        """
        Retrieve a list of TicketSearchResult objects for the given optional search arguments

        Method: GET
        URL: /securechangeworkflow/api/securechange/tickets/search_by/group
        Version: R23-1+

        Usage:
            # Example #1:
            search_results = scw.ticket_search_by_group(group_id=9)
        """
        from pytos2.securechange.ticket import TicketSearchResult

        params = {
            "group_id": group_id,
        }

        r = self.api.session.get("tickets/search_by/group", params=params)
        r_json = self.api.handle_json(r, "ticket_search_by_free_text", action="get")
        tickets = get_api_node(
            r_json, "tickets_search_results.ticket_result", listify=True
        )

        return [TicketSearchResult.kwargify(t) for t in tickets]

    def ticket_search_by_free_text(
        self, parameter: str
    ) -> List["pytos2.securechange.ticket.TicketSearchResult"]:
        """
        Retrieve a list of TicketSearchResult objects for the given optional search arguments

        Method: GET
        URL: /securechangeworkflow/api/securechange/tickets/search_by/free_text
        Version: R23-1+

        Usage:
            # Example #1:
            search_results = scw.ticket_search_by_free_text(parameter = "taskstatus:ASSIGNED")

        Parameter reference URL: https://forum.tufin.com/support/kc/R24-2/Content/Suite/2331.htm
        """
        from pytos2.securechange.ticket import TicketSearchResult

        params = {
            "parameter": parameter,
        }

        r = self.api.session.get("tickets/search_by/free_text", params=params)
        r_json = self.api.handle_json(r, "ticket_search_by_free_text", action="get")
        tickets = get_api_node(
            r_json, "tickets_search_results.ticket_result", listify=True
        )

        return [TicketSearchResult.kwargify(t) for t in tickets]

    def get_users(
        self,
        show_indirect_relation: Optional[bool] = None,
        user_name: Optional[str] = None,
        email: Optional[str] = None,
        exact_name: Optional[bool] = None,
    ) -> List[SCWParty]:
        """
        Retrieve a list of SCWParty objects for the given optional search arguments

        Method: GET
        URL: /securechangeworkflow/api/securechange/users
        Version: R22-2+

        Usage:
            # Example #1:
            users = scw.get_users(user_name='ateam')

            # Example #2:
            users = scw.get_users(email='ateam@tufin.com')
        """
        params = {}
        if show_indirect_relation:
            params["showIndirectRelation"] = show_indirect_relation
        if user_name:
            params["user_name"] = user_name
        if email:
            params["email"] = email
        if exact_name:
            params["exact_name"] = exact_name

        response = self.api.session.get("users", params=params)
        if not response.ok:
            response.raise_for_status()
        else:
            _json = response.json()
            users_node = get_api_node(_json, "users.user", listify=True)

            users = []
            self.user_cache.clear()

            for obj in users_node:
                user = classify_user_object(obj)
                users.append(user)

            self.user_cache.set_data(users)
            return users

    def _get_user_from_server(self, identifier: int) -> SCWParty:
        """
        Retrieve a SCWParty object for the given id

        Method: GET
        URL: /securechangeworkflow/api/securechange/users/{id:[0-9]+}
        Version: 22-2+

        Usage:
            user = st._get_user_from_server(5)
        """
        response = self.api.session.get(f"users/{identifier}")

        # if not response.ok:
        response.raise_for_status()
        # else:
        _json = response.json()

        key = ""
        if "group" in _json:
            key = "group"
        elif "user" in _json:
            key = "user"
        else:
            raise KeyError(
                f"Root user class key {_json.keys()} not currently supported by pytos2"
            )

        user_json = _json[key]
        if isinstance(user_json, list):
            user_json = user_json[0]
        return classify_user_object(user_json, obj_type=key)

    def get_user(
        self,
        identifier: Union[str, int],
        expand: bool = False,
        update_cache: Optional[bool] = None,
    ) -> SCWParty:
        """
        Retrieve a SCWParty object for the given id, optionally caching or inclusion of @xsi.type

        Method: GET
        URL: /securechangeworkflow/api/securechange/users (if update_cache is not False)
        URL: /securechangeworkflow/api/securechange/users/{id:[0-9]+} (for including @xsi.type)
        Version: 22-2+

        Usage:
            # Example #1:
            user = scw.get_user(5)

            # Example #2:
            user = scw.get_user(5, expand=True)
        """
        if update_cache is not False and self.user_cache.is_empty():
            _ = self.get_users()  # create or update cache

        if isinstance(identifier, str):
            user = self.users_by_name.get(identifier)
            if not user:
                _ = self.get_users()
                user = self.users_by_name.get(identifier)
                if not user:
                    raise ValueError(f"User with name {identifier} not found")

            if expand:
                # this API only give @xsi.type as additional info
                return self._get_user_from_server(user.id)
            else:
                return user
        else:
            user = self.users_by_id.get(identifier)
            if not user:
                _ = self.get_users()
                user = self.users_by_id.get(identifier)
            if not user:
                raise ValueError(f"User with id {identifier} not found")

            if expand:
                try:
                    return self._get_user_from_server(identifier)
                except HTTPError as e:
                    # wrap the HTTPError into ValueError for consisency
                    raise ValueError(f"User with id {identifier} not found got {e}")
            else:
                return user

    def add_user_group(self, name: str, description: str, email: str) -> SCWGroup:
        """
        Method: POST
        URL: /securechangeworkflow/api/securechange/users/group
        Version: R22-1+

        Usage:
            scw.add_user_group("Group1", "Test Group 1", "group1@contoso.com")
        """
        post_data = {
            "group": {"name": name, "description": description, "email": email}
        }
        response = self.api.session.post("users/group", json=post_data)
        return self.api.handle_creation(response, "add_user_group", cls=SCWGroup)

    def update_user_group(
        self,
        group_id: int,
        member_ids_to_add: Optional[List[int]] = None,
        member_ids_to_remove: Optional[List[int]] = None,
    ) -> None:
        """
        Method: PUT
        URL: /securechangeworkflow/api/securechange/users/group/{id:[0-9]+}
        Version: R22-1+

        Usage:
            scw.update_user_group(54, ["70","96"], ["85","49"])
        """
        put_data = {
            "group_change": {
                "member_ids_to_add": member_ids_to_add,
                "member_ids_to_remove": member_ids_to_remove,
            }
        }
        response = self.api.session.put(f"users/group/{group_id}", json=put_data)
        self.api.handle_response(response, "update_user_group", "update")

    def delete_user_or_group(self, identifier: int) -> None:
        """
        URL: /securechangeworkflow/api/securechange/users/{id:[0-9]+}
        Version: R22-1+
        Notes:
            Delete a single user or group from SecureChange Users

        Usage:
            scw.delete_user_or_group(70)
        """
        response = self.api.session.delete(f"users/{identifier}")
        self.api.handle_response(response, "delete_user_or_group", "delete")

    def user_import_ldap(self, ldap: LDAPParty) -> None:
        """
        URL: /securechangeworkflow/api/securechange/users/import
        Version: R22-1+
        Notes:
            Import a single user or group using the LDAP already configured in SecureChange

        Usage:
            ldap = LDAPParty()
            ldap.ldap_configuration_name = "Contoso LDAP" #The name you have configured in the UI under SecureChange LDAP
            ldap.ldap_dn = "CN=abaker,OU=Users,DC=tcse,DC=net"
            scw.user_import_ldap(ldap)
        """
        post_data = {"ldap_party": ldap._json}
        response = self.api.session.post("users/import", json=post_data)
        self.api.handle_response(response, "user_import_ldap", "create/import")

    def user_login(self, user: LDAPUserSimulateLogin):
        """
        This API implicitly imports LDAP users as automatics users to SecureChange by simulating their login.
        To explicitly import LDAP entities, both users and groups, it is recommended to use /users/import API

        URL: /securechangeworkflow/api/securechange/users
        Version: R22-1+
        Notes:
            ldap_configuration is optional and if not supplied, the default ldap config will be used automatically

        Usage:
            user = LDAPUserSimulateLogin()
            user.name = 'abaker'
            ldap_config = LDAPConfiguration()
            ldap_config.name = "Contoso LDAP"
            ldap_config.id = 1
            user.ldap_configuration = ldap_config
            scw.user_login(user)

        """
        if not isinstance(user, LDAPUserSimulateLogin):
            raise TypeError("user argument must be of type LDAPUserSimulateLogin")

        post_data = {"user": user._json}
        response = self.api.session.post("users", json=post_data)
        self.api.handle_response(response, "user_login", "create")

    def add_user(self, new_user: NewLocalUser):
        """
        URL: /securechangeworkflow/api/securechange/users
        Version: R22-1+

        Usage:
            new_user = NewLocalUser()
            new_user.name = 'abaker'
            new_user.first_name = "Aaron"
            new_user.last_name = "Baker"
            new_user.email = "aaron.baker@contoso.com"
            new_user.password = "password"
            new_user.notes = "asdf"
            new_user.default_authentication = False
            scw.user_create(new_user)

        """
        post_data = {"localuser": new_user._json}
        response = self.api.session.post("users", json=post_data)
        return self.api.handle_creation(response, "add_user", cls=SCWUser)

    def group_login(self, group: LDAPSimulateLogin):
        """
        This API implicitly imports LDAP users as automatics users to SecureChange by simulating their login.
        To explicitly import LDAP entities- both users and groups, it is recommended to use /users/import API

        URL: /securechangeworkflow/api/securechange/users
        Version: R22-1+
        Notes:
            ldap_configuration is optional and if not supplied, the default ldap config will be used automatically

        Usage:
            group = LDAPSimulateLogin()
            group.name = 'Accountants'
            ldap_config = LDAPConfiguration()
            ldap_config.name = "Contoso LDAP"
            ldap_config.id = 1
            group.ldap_configuration = ldap_config
            scw.group_login(group)

        """
        if not isinstance(group, LDAPSimulateLogin):
            raise TypeError("group argument must be of type LDAPSimulateLogin")

        post_data = {"group": group._json}
        response = self.api.session.post("users", json=post_data)
        self.api.handle_response(response, "group_login", "import")

    def get_excluded_device_ids(self, show_all: Optional[bool] = None) -> List[int]:
        """
        Retrieve a List of ints that make up the list of excluded device ids

        Method: GET
        URL: /securechangeworkflow/api/securechange/devices/excluded
        URL: /securechangeworkflow/api/securechange/devices/excluded?show_all=true
        Version: 22-2+

        Usage:
            # Example #1:
            excluded_device_ids = scw.get_excluded_device_ids()

            # Example #2:
            excluded_device_ids = scw.get_excluded_device_ids(show_all=True)
        """
        url = "devices/excluded"
        if isinstance(show_all, bool):
            url += "?show_all="
            show_all = str(show_all).lower()
            url += show_all

        r = self.api.session.get(url)
        if r.ok:
            excludes_json = r.json()
            device_ids_model = DeviceExclusions.kwargify(excludes_json)
            return device_ids_model.device_ids
        else:
            r.raise_for_status()

    def get_tickets(
        self,
        status: Optional[
            Union[
                "pytos2.securechange.ticket.TicketStatus",
                List["pytos2.securechange.ticket.TicketStatus"],
            ]
        ] = None,
        start: Optional[int] = None,
        descending: Optional[bool] = None,
        expand_links: Optional[bool] = None,
    ) -> List["pytos2.securechange.ticket.TicketIterator"]:
        """
        Retrieve a list of TicketIterator objects for the given optional search arguments

        Method: GET
        URL: /securechangeworkflow/api/securechange/tickets
        Version: R22-2+

        Usage:
            # Example #1:
            ticket_iterators = scw.get_tickets()

            # Example #2:
            ticket_iterators = scw.get_users(descending=True)
        """
        from pytos2.securechange.ticket import TicketStatus, TicketIterator

        params = {}

        if status is not None:
            if isinstance(status, list):
                params["status"] = ",".join([s.value for s in status])
            else:
                params["status"] = status.value
        if start is not None:
            params["start"] = start
        if descending is not None:
            params["desc"] = "true" if descending else "false"
        if expand_links is not None:
            params["expand_links"] = "true" if expand_links else "false"

        return list(TicketIterator(self.api.session, params))

    def get_ticket(self, _id: int) -> "pytos2.securechange.ticket.Ticket":
        """
        Retrieve a Ticket object for the given id

        Method: GET
        URL: /securechangeworkflow/api/securechange/tickets/{id:[0-9]+}
        Version: 22-2+

        Usage:
            ticket = scw.get_ticket(5)
        """
        from pytos2.securechange.ticket import Ticket

        r = self.api.session.get(f"tickets/{_id}")
        if r.ok:
            tkt = Ticket.kwargify(r.json())
            return tkt
        else:
            r.raise_for_status()

    def change_requester(
        self,
        ticket: Union[int, "pytos2.securechange.ticket.Ticket"],
        user: Union[str, int, SCWParty],
        comment="",
    ) -> None:
        """
        Change ticket requester.

        Method: PUT
        URL: /securechangeworkflow/api/securechange/tickets/{id:[0-9]+}/change_requester/{assigneeId:[0-9]+}
        Version: 22-2+

        Usage:
            # Example #1:
            ticket = scw.get_ticket(1)
            user = scw.get_user(5)
            scw.change_requester(ticket, user, 'change requester')

            # Example #2:
            scw.change_requester(1, 5, 'change requester')
        """
        from pytos2.securechange.ticket import Ticket

        if not isinstance(ticket, Ticket):
            ticket = self.get_ticket(ticket)
        if not isinstance(user, SCWUser):
            user = self.get_user(user)

        if isinstance(user, SCWGroup):
            raise ValueError(
                "Error changing requester: Groups not allowed as requesters"
            )

        try:
            url = f"tickets/{ticket.id}/change_requester/{user.id}"
            put_body = {"comment": {"comment": comment if comment else "comment"}}
            response = self.api.session.put(url, json=put_body)
            response.raise_for_status()
        except RequestException as e:
            raise ValueError(f"Error changing requester: {e.response.text}")

    def reassign_ticket(
        self,
        ticket,
        user,
        step: Union[None, "pytos2.securechange.ticket.Step", int, str] = None,
        task: Union[None, "pytos2.securechange.ticket.Task", int] = None,
        comment="",
    ) -> None:
        """
        Reassign a ticket task to the given user, the task defaulting to the last task of the current step.
        Optionally a speficied step or task can be passed as an argument.

        Method: PUT
        URL: /securechangeworkflow/api/securechange/tickets/{id:[0-9]+}/steps/{stepId:[0-9]+}/tasks/{taskId:[0-9]+}/reassign/{assigneeId:[0-9]+}
        Payload:
            {
                "reassign_task_comment": {
                    "comment": comment
                }
            }
        Version: 22-2+

        Usage:
            scw.reassign_ticket(ticket, user)
        """
        from pytos2.securechange.ticket import Step, Task, Ticket

        if not isinstance(ticket, Ticket):
            ticket = self.get_ticket(ticket)

        if not isinstance(user, SCWUser):
            user = Scw.default.get_user(user)

        if step is None:
            step = ticket.current_step

        if not isinstance(step, Step):
            step = ticket.get_step(step)

        if task is None:
            task = step.get_task(0)

        if not isinstance(task, Task):
            task = step.get_task(task)

        try:
            response = self.api.session.put(
                f"tickets/{ticket.id}/steps/{step.id}/tasks/{task.id}/reassign/{user.id}",
                json={"reassign_task_comment": {"comment": comment}},
            )
            if not response.ok:
                msg = response.json().get("result").get("message")
                response.raise_for_status()
        except HTTPError as e:
            raise ValueError(
                f"Got {e}, with Error Message: {msg}. Only tasks under current step can be reassigned"
            )

    def get_attachment(self, file_id: str) -> bytes:
        """
        Retrieve an attachment file's contents (bytes) for the given file_id

        Method: GET
        URL: /securechangeworkflow/api/securechange/attachments/{uid}
        Version: 22-2+

        Usage:
            file_contents = scw.get_attachment('d95d574f-49c0-4c91-b08f-b243735a7501')
        """
        response = self.api.session.get(f"attachments/{file_id}")
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Getting Attachment: {e}")
        return response.content

    def add_attachment(self, file: str) -> str:
        """
        Post the contents of the file with the given filename as an attachment

        Method: POST
        URL: /securechangeworkflow/api/securechange/attachments
        Version: 22-2+

        Usage:
            scw.add_attachment('/conf/attachment_file.csv')
        """
        attachment = {"attachment": open(file, "rb")}
        response = self.api.session.post("attachments", files=attachment)
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Attachment: {e}")
        return response.text

    def add_comment(
        self,
        ticket_id: Union[int, str],
        step_id: Union[int, str],
        task_id: Union[int, str],
        comment_content: str,
        attachment_uuids: Optional[List[str]],
    ) -> str:
        """
        Post the the given comment_contents and, optionally, file attachments specified by attachment_uuids
        as an attachment to the given ticket, step and task.

        Method: POST
        URL: /securechangeworkflow/api/securechange/attachments
        Version: 22-2+

        Usage:
            # Example #1:
            scw.add_comment(100, 200, 201, 'Thie is a comment')

            # Example #2:
            scw.add_comment(
                100, 200, 201, 'Thie is a comment with an attachment', ['d95d574f-49c0-4c91-b08f-b243735a7501']
            )
        """

        def _format_attachments(uuids: List[str]) -> dict:
            return {"attachment": [{"uid": id} for id in uuids]}

        comment = {"comment": {"content": comment_content}}
        if attachment_uuids:
            comment["comment"]["attachments"] = _format_attachments(attachment_uuids)

        response = self.api.session.post(
            f"tickets/{ticket_id}/steps/{step_id}/tasks/{task_id}/comments",
            json=comment,
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Adding Comment: {e}")

        return response.text

    def delete_comment(
        self,
        ticket_id: Union[int, str],
        comment_id: Union[int, str],
    ) -> None:
        """
        Delete the ticket comment for the given ticket_id, comment_id

        Method: DELETE
        URL: /securechangeworkflow/api/securechange/tickets/{ticketId:[0-9]+}/comments/{commentId:[0-9]+}
        Version: 22-2+

        Usage:
            scw.delete_comment(100, 1)
        """
        response = self.api.session.delete(
            f"tickets/{ticket_id}/comments/{comment_id}",
        )
        if not response.ok:
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ValueError(f"Error Deleting Comment: {e}")

    def get_ticket_history(
        self, _id: int
    ) -> "pytos2.securechange.ticket.TicketHistory":
        """
        Retrieve a TicketHistory object for the given ticket id

        Method: GET
        URL: /securechangeworkflow/api/securechange/tickets/{id:[0-9]+}/history
        Version: 22-2+

        Usage:
            # Example #1:
            ticket_history = scw.get_ticket_history(5)

            # Example #2:
            ticket = scw.get_ticket(5)
            ticket_history = ticket.get_history()
        """
        from pytos2.securechange.ticket import TicketHistory

        try:
            response = self.api.session.get(f"tickets/{_id}/history")
            response.raise_for_status()
            response_json = response.json()
            ticket_history = TicketHistory.kwargify(
                response_json["ticket_history_activities"]
            )
            return ticket_history
        except JSONDecodeError:
            raise ValueError(f"Error decoding json: {response.text}")
        except ValueError as e:
            raise ValueError(f"Error creating ticket_history class: {e}")
        except RequestException as e:
            raise ValueError(f"Error retrieving ticket_history: {e}")

    def get_triggers(self) -> List[WorkflowTrigger]:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/triggers
        Version: R22-2+

        Usage:
            triggers = scw.get_triggers()

        """
        try:
            url = "triggers"
            response = self.api.session.get(url)
            response.raise_for_status()
            response_json = response.json()
            api_nodes = get_api_node(
                response_json, "workflow_triggers.workflow_trigger", listify=True
            )
            return [WorkflowTrigger.kwargify(trigger) for trigger in api_nodes]
        except JSONDecodeError as e:
            raise ValueError(f"Error decoding json: {response.text}") from e
        except ValueError as e:
            raise ValueError(f"Error creating triggers class: {e}") from e
        except RequestException as e:
            raise ValueError(f"Error retrieving triggers: {e}") from e

    def get_trigger(
        self, name: Optional[str] = None, id: Optional[int] = None
    ) -> WorkflowTrigger:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/triggers
        Version: R22-2+

        Usage:
            # Example #1:
            trigger = scw.get_trigger(name='trigger name')

           # Example #2:
            trigger = scw.get_trigger(id=1000)
        """
        if (name and id) or not (name or id):
            raise ValueError("Either name or id (but not both) must be supplied")
        query = f"name={name}" if name else f"id={id}"

        try:
            url = f"triggers?{query}"
            response = self.api.session.get(url)
            response.raise_for_status()
            response_json = response.json()
            api_nodes = get_api_node(
                response_json, "workflow_triggers.workflow_trigger", listify=True
            )
            return WorkflowTrigger.kwargify(api_nodes[0])
        except JSONDecodeError as e:
            raise ValueError(f"Error decoding json: {response.text}") from e
        except ValueError as e:
            raise ValueError(f"Error creating triggers class: {e}") from e
        except RequestException as e:
            raise ValueError(f"Error retrieving trigger: {e.response.text}") from e

    def add_trigger(self, workflow_trigger: WorkflowTrigger) -> None:
        """
        Method: POST
        URL: /securechangeworkflow/api/securechange/triggers
        Version: R22-2+

        Usage:
            script = Script(path='test-solution/test-solution', arguments='')
            workflow = Workflow(name='test workflow', parent_workflow_id=1)
            trigger = Trigger(name='test trigger', workflow=workflow, events=['CREATE', 'ADVANCE'])
            workflow_trigger = WorkflowTrigger(name='test', script=script, triggers=[trigger])
            scw.add_trigger(workflow_trigger)
        """
        try:
            url = "triggers"
            json_data = {
                "workflow_triggers": {"workflow_trigger": [workflow_trigger._json]}
            }
            response = self.api.session.post(url, json=json_data)
            response.raise_for_status()
        except HTTPError as e:
            raise ValueError(f"Error posting trigger: {e.response.text}") from e
        except RequestException as e:
            raise ValueError("Error posting trigger") from e

    def get_ticket_events(
        self,
        type: Optional[str] = None,
        assignee_name: Optional[int] = None,
        assignee_id: Optional[int] = None,
        participant_name: Optional[str] = None,
        participant_id: Optional[int] = None,
        ticket_id: Optional[int] = None,
        workflow_name: Optional[str] = None,
        parent_workflow_id: Optional[int] = None,
        step_name: Optional[str] = None,
        step_id: Optional[int] = None,
        date_from: DateFilterType = None,
        date_to: DateFilterType = None,
    ) -> Iterable["pytos2.securechange.ticket.TicketEvent"]:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/tickets/lifecycle_events
        Version: R24-1+

        Usage:
            # Example #1:
            ticket_events = scw.get_ticket_events()

            # Example #2:
            ticket_events = scw.get_ticket_events(ticket_id=1031)
        """
        from pytos2.securechange.ticket import TicketEvent

        url = "tickets/lifecycle_events"
        params = {}
        if type:
            params["type"] = type
        if assignee_name:
            params["assignee_name"] = assignee_name
        if assignee_id:
            params["assignee_id"] = assignee_id
        if participant_name:
            params["participant_name"] = participant_name
        if participant_id:
            params["participant_id"] = participant_id
        if ticket_id:
            params["ticket_id"] = ticket_id
        if workflow_name:
            params["workflow_name"] = workflow_name
        if parent_workflow_id:
            params["parent_workflow_id"] = parent_workflow_id
        if step_name:
            params["step_name"] = step_name
        if step_id:
            params["step_id"] = step_id
        if date_from:
            params["date_from"] = coerce_date_filter_type(date_from)
        if date_to:
            params["date_to"] = coerce_date_filter_type(date_to)
        api_node = "ticket_events.ticket_event"
        pager = Pager(
            self.api,
            url,
            api_node,
            "get_ticket_events",
            TicketEvent.kwargify,
            params=params,
        )
        return pager

    def backfill_ticket_events(self, from_date: str) -> None:
        """
        Method: POST
        URL: /securechangeworkflow/api/securechange/tickets/lifecycle_events/historical_events
        Version: R24-1+

        Usage:
            scw.backfill_ticket_events("2023-11-05")
        """
        try:
            url = "tickets/lifecycle_events/historical_events"
            json_data = {"ticket_events_create_history_request": {"from": from_date}}
            response = self.api.session.post(url, json=json_data)
            response.raise_for_status()
        except (HTTPError, ConnectionError, Timeout, RequestException) as e:
            raise ValueError(f"Error retrieving ticket_historical_events_status: {e}")

    def get_ticket_historical_events_status(self) -> str:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/tickets/lifecycle_events/historical_events_status
        Version: R24-1+

        Usage:
            historical_events_status = scw.get_ticket_historical_events_status()
        """
        try:
            url = "tickets/lifecycle_events/historical_events_status"
            response = self.api.session.get(url)
            response.raise_for_status()
            response_json = response.json()
            return response_json["ticket_events_created_history_status"]["run_status"]
        except (HTTPError, ConnectionError, Timeout, RequestException) as e:
            raise ValueError(f"Error retrieving ticket_historical_events_status: {e}")

    def map_rules(self, ticket_id: int, handler_id: int = None) -> None:
        """
        Method: POST
        URL: /securechangeworkflow/api/securechange/tickets/{id:[0-9]+}/map_rules

        Version: R24-1+

        Usage:
            # Example:
            scw.map_rules(1037)
        """
        try:
            url = f"tickets/{ticket_id}/map_rules"
            if handler_id:
                url += f"?handler_id={handler_id}"
            response = self.api.session.post(url)
            response.raise_for_status()
        except (HTTPError, ConnectionError, Timeout, RequestException) as e:
            raise ValueError(
                f"Error running map_rules: {e} response: {e.response.text}"
            )

    def designer_redesign(
        self, ticket_id: int, step_id: int, task_id: int, field_id: int
    ) -> None:
        """
        Method: PUT
        URL: /securechangeworkflow/api/securechange/tickets/{id:[0-9]+}/steps/{stepId:[0-9]+}/tasks/{taskId:[0-9]+}/fields/{fieldId:[0-9]+}/designer/redesign

        Version: R24-1+

        Usage:
            # Example:
            scw.designer_redesign(1037, 5689, 5705, 61671)
        """
        try:
            url = f"tickets/{ticket_id}/steps/{step_id}/tasks/{task_id}/fields/{field_id}/designer/redesign"
            response = self.api.session.put(url)
            response.raise_for_status()
        except (HTTPError, ConnectionError, Timeout, RequestException) as e:
            raise ValueError(
                f"Error running designer redesign: {e} response: {e.response.text}"
            )

    def designer_device_commit(
        self, ticket_id: int, step_id: int, task_id: int, field_id: int, device_id: int
    ) -> None:
        """
        Method: PUT
        URL: /securechangeworkflow/api/securechange/tickets/{id:[0-9]+}/steps/{stepId:[0-9]+}/tasks/{taskId:[0-9]+}/fields/{fieldId:[0-9]+}/designer/device/{deviceId:[0-9]+}/commit

        Version: R24-1+

        Usage:
            # Example:
            scw.designer_device_commit(1037, 5689, 5705, 61671, 8)
        """
        try:
            url = f"tickets/{ticket_id}/steps/{step_id}/tasks/{task_id}/fields/{field_id}/designer/device/{device_id}/commit"
            response = self.api.session.put(url)
            response.raise_for_status()
        except (HTTPError, ConnectionError, Timeout, RequestException) as e:
            raise ValueError(
                f"Error running designer device commit: {e} response: {e.response.text}"
            )

    def designer_device_update(
        self,
        ticket_id: int,
        step_id: int,
        task_id: int,
        field_id: int,
        device_id: int,
        force: Optional[bool] = False,
    ) -> None:
        """
        Method: PUT
        URL: /securechangeworkflow/api/securechange/tickets/{id:[0-9]+}/steps/{stepId:[0-9]+}/tasks/{taskId:[0-9]+}/fields/{fieldId:[0-9]+}/designer/device/{deviceId:[0-9]+}/update

        Version: R24-1+

        Usage:
            # Example #1:
            scw.designer_device_update(1037, 5689, 5705, 61671, 8)

            # Example #2:
            scw.designer_device_update(1037, 5689, 5705, 61671, 8, True)
        """
        try:
            url = f"tickets/{ticket_id}/steps/{step_id}/tasks/{task_id}/fields/{field_id}/designer/device/{device_id}/update"
            if force:
                url += "?force=true"
            response = self.api.session.put(url)
            response.raise_for_status()
        except (HTTPError, ConnectionError, Timeout, RequestException) as e:
            raise ValueError(
                f"Error running designer device update: {e} response: {e.response.text}"
            )

    def get_workflows(
        self, workflow_type: Optional[WorkflowType] = None
    ) -> List[BasicWorkflowInfo]:
        """
        URL: /securechangeworkflow/api/securechange/workflows/active_workflows
        URL: /securechangeworkflow/api/securechange/workflows/active_workflows?type=ACCESS_REQUEST
        Version: R21-3+

        Usage:
            workflows = scw.get_workflows()
            workflows = scw.get_workflows(WorkflowType.ACCESS_REQUEST)

        """

        try:
            url = (
                f"workflows/active_workflows?type={workflow_type.value}"
                if workflow_type
                else "workflows/active_workflows"
            )
            response = self.api.session.get(url)
            response.raise_for_status()
            response_json = response.json()
            return [
                BasicWorkflowInfo.kwargify(wf)
                for wf in get_api_node(
                    response_json, "workflows.workflow", listify=True
                )
            ]
        except JSONDecodeError as e:
            raise ValueError(f"Error decoding json: {response.text}") from e
        except RequestException as e:
            raise ValueError(f"Error retrieving workflows: {e}") from e

    def get_workflow(
        self,
        workflow_id: Optional[Union[int, str]] = None,
        workflow_name: Optional[str] = None,
    ) -> FullWorkflow:
        """
        This function returns a FullWorkflow object or a sub-type of it. If something you need is
        not mapped in FullWorkflow, you may use the `data' attribute to access it.

        URL: /securechangeworkflow/api/securechange/workflows?id=588
        URL: /securechangeworkflow/api/securechange/workflows?name=Contoso%20Workflow
        Version:

        Usage:
            workflow = scw.get_workflow(588)
            workflow = scw.get_workflow(workflow_name="Contoso Workflow")

        """
        try:
            url = (
                f"workflows?id={workflow_id}"
                if workflow_id
                else f"workflows?name={workflow_name}"
            )
            response = self.api.session.get(url)
            response.raise_for_status()
            # Returning raw dictionary because we do not yet have models for the entire workflow
            # This is because the name of the first key varies depending on the workflow type
            workflow_data = response.json()
            return FullWorkflow.kwargify(workflow_data)

            # workflow_data = response_json[list(response_json.keys())[0]]
            # return BasicWorkflowInfo.kwargify(workflow_data)
        except JSONDecodeError as e:
            raise ValueError(f"Error decoding json: {response.text}") from e
        except RequestException as e:
            raise ValueError(f"Error retrieving workflow by name or id: {e}") from e

    def get_extensions(self) -> List[MarketplaceApp]:
        """
        URL: /securechangeworkflow/api/securechange/extensions
        Version: R23-2+

        Usage:
            apps = scw.get_extensions()

        """

        try:
            response = self.api.session.get("extensions")
            response.raise_for_status()
            response_json = response.json()
            return [
                MarketplaceApp.kwargify(ext)
                for ext in get_api_node(response_json, "extensions")
            ]
        except JSONDecodeError as e:
            raise ValueError(f"Error decoding json: {response.text}") from e
        except RequestException as e:
            raise ValueError(f"Error retrieving extensions/marketplaceapps: {e}") from e

    def get_roles(self) -> List[UserRole]:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/roles
        Version: R22-2+

        Usage:
            roles = scw.get_roles()

        """
        response = self.api.session.get("roles")
        json = self.api.handle_json(response, "get_roles", "get")
        return [
            UserRole.kwargify(role)
            for role in get_api_node(json, "roles.role", listify=True)
        ]

    def get_role(self, id: Union[int, str]) -> UserRole:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/roles/{id:[0-9]+}
        Version: R22-2+

        Usage:
            role = scw.get_role(8)

        """
        response = self.api.session.get(f"roles/{id}")
        json = self.api.handle_json(response, "get_role", "get")
        return UserRole.kwargify(json["role"])

    def update_user_roles(self, user_id: int, roles_list: List[UserUpdateRole]):
        """
        Method: PUT
        URL: /securechangeworkflow/api/securechange/users/{id:[0-9]+}/roles
        Version: R21-3+
        Notes:
            Clears existing roles and updates with the new list, except Requestor role will always be present
            User get_user to see what Roles the user is assigned

        Usage:
            roles = []
            roles.append(UserUpdateRole(name="Auditor"))
            sc.user_update_roles(70, roles)

        """
        json_roles = [user_role._json for user_role in roles_list]
        post_data = {"roles": {"role": json_roles}}
        response = self.api.session.put(f"users/{user_id}/roles", json=post_data)
        self.api.handle_response(response, "user_update_roles", "update")

    def get_saved_searches(self) -> List[Query]:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/queries
        Version: R24-1+

        Usage:
            # Example #1:
            queries = scw.get_saved_searches()
        """

        try:
            url = "queries"

            response = self.api.session.get(url)
            response.raise_for_status()
            response_json = response.json()

            queries = [
                Query.kwargify(query)
                for query in get_api_node(response_json, "queries.query", listify=True)
            ]

            return queries

        except (HTTPError, ConnectionError, Timeout, RequestException) as e:
            raise ValueError(
                f"Error retrieving queries: {e} response: {e.response.text}"
            )

    def get_saved_search(self, id: int) -> List[Query]:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/queries/{id:[0-9]+}
        Version: R24-1+

        Usage:
            # Example #1:
            queries = scw.get_saved_search(id=14)
        """

        try:
            url = f"queries/{id}"
            response = self.api.session.get(url)
            response.raise_for_status()
            response_json = response.json()
            queries = []

            if response_json["query"]["type"] == "DETAILED":
                queries = [
                    DetailedQuery.kwargify(query)
                    for query in get_api_node(response_json, "query", listify=True)
                ]

            if response_json["query"]["type"] == "FREE_TEXT":
                queries = [
                    FreeTextQuery.kwargify(query)
                    for query in get_api_node(response_json, "query", listify=True)
                ]

            return queries

        except (HTTPError, ConnectionError, Timeout, RequestException) as e:
            raise ValueError(
                f"Error retrieving queries: {e} response: {e.response.text}"
            )

    def add_saved_search(self, query: Query) -> None:
        """
        Method: POST
        URL: /securechangeworkflow/api/securechange/queries
        Version: R24-1+

        Usage:
            # Example #1:
            queries = scw.add_saved_search(query)
        """
        try:
            url = "queries/"
            response = self.api.session.post(url, json=query)
            response.raise_for_status()
            return response.status_code
        except (HTTPError, ConnectionError, Timeout) as e:
            raise ValueError(f"Error deleting query: {e} response: {e.response.text}")
        except JSONDecodeError as e:
            raise ValueError(f"Error decoding json: {response.text}") from e
        except RequestException as e:
            raise ValueError(f"Error retrieving saved searches: {e}") from e

    def update_saved_search(self, id, query: Query) -> None:
        """
        Method: PUT
        URL: /securechangeworkflow/api/securechange/queries/{id}
        Version: R24-1+

        Usage:
            # Example #1:
            queries = scw.update_saved_search(id, query)
        """
        try:
            url = "queries/" + str(id)
            response = self.api.session.put(url, json=query)
            response.raise_for_status()
            return response.status_code
        except (HTTPError, ConnectionError, Timeout) as e:
            raise ValueError(f"Error deleting query: {e} response: {e.response.text}")
        except JSONDecodeError as e:
            raise ValueError(f"Error decoding json: {response.text}") from e
        except RequestException as e:
            raise ValueError(f"Error retrieving saved searches: {e}") from e

    def delete_saved_search(self, id) -> None:
        """
        Method: DELETE
        URL: /securechangeworkflow/api/securechange/queries/{id}
        Version: R24-1+

        Usage:
            # Example #1:
            queries = scw.delete_saved_search(id)
        """
        try:
            url = "queries/" + str(id)
            response = self.api.session.delete(url)
            response.raise_for_status()
            return response.status_code
        except (HTTPError, ConnectionError, Timeout) as e:
            raise ValueError(f"Error deleting query: {e} response: {e.response.text}")
        except JSONDecodeError as e:
            raise ValueError(f"Error decoding json: {response.text}") from e
        except RequestException as e:
            raise ValueError(f"Error retrieving saved searches: {e}") from e

    def get_domains(self) -> List[Domain]:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/domains
        Version: R21-1+

        Usage:
            domains = scw.get_domains()

        """
        response = self.api.session.get("domains")
        json = self.api.handle_json(response, "get_domains", "get")
        return [
            Domain.kwargify(domain)
            for domain in get_api_node(json, "domains.domain", listify=True)
        ]

    def get_domain(self, _id: Union[int, str]) -> Domain:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/domains/{domainId:[0-9]+}
        Version: R21-1+

        Usage:
            domain = scw.get_domain(1)

        """
        response = self.api.session.get(f"domains/{_id}")
        json = self.api.handle_json(response, "get_domain", "get")
        return Domain.kwargify(get_api_node(json, "domain"))

    def synchronize_domains(self):
        """
        Method: PUT
        URL: /securechangeworkflow/api/securechange/domains/synchronize
        Version: R21-1+

        Usage:
            scw.synchronize_domains()

        """
        response = self.api.session.put("domains/synchronize")
        self.api.handle_response(response, "synchronize_domains", "update")

    def get_requests(
        self,
        view: str,
        group_id: Optional[int] = None,
        request_id: Optional[int] = None,
        type: Optional[str] = None,
        ticket_status: Optional[List[str]] = None,
        ticket_id: Optional[int] = None,
        ticket_requires_attention: Optional[bool] = None,
        logical_operator: Optional[str] = None,
    ) -> List[RequestsSearchList]:
        """
        Method: GET
        URL: /securechangeworkflow/api/securechange/requests/search
        Version: R23-2+

        Usage:
            requests = scw.get_requests(view="ALL_REQUESTS")
            requests = scw.get_requests(view="ALL_REQUESTS",ticketId=815)

        """
        url = "requests/search"
        params = {}
        if view:
            params["view"] = view
        if group_id:
            params["groupId"] = group_id
        if request_id:
            params["requestId"] = request_id
        if type:
            params["type"] = type
        if ticket_status:
            params["ticketStatus"] = ticket_status
        if ticket_id:
            params["ticketId"] = ticket_id
        if ticket_requires_attention:
            params["ticketRequiresAttention"] = ticket_requires_attention
        if logical_operator:
            params["logicalOperator"] = logical_operator

        response = self.api.session.get(url, params=params)
        response_json = self.api.handle_json(response, "get_requests", "get")
        requests_list = get_api_node(response_json, "requests", listify=True)

        queries = [
            (
                TicketRequestSearch.kwargify(request)
                if request["type"] == "TICKET"
                else TicketDraftRequestSearch.kwargify(request)
            )
            for request in requests_list
        ]

        return queries

    def cancel_request(self, id) -> None:
        """
        Method: PUT
        URL: /securechangeworkflow/api/securechange/requests/{id:[0-9]+}/cancel
         Version: R23-2+

        Usage:
            # Example #1:
            request = scw.cancel_request(id)
        """
        url = "requests/" + str(id) + "/cancel"
        response = self.api.session.put(url)
        self.api.handle_response(response, "cancel_request", "cancel")



from enum import Enum

import copy


class CacheEvent(Enum):
    ADD = "add"
    REMOVE = "remove"
    CLEAR = "clear"
    SET = "set"


class Cache:
    def __init__(self):
        self._cache = []
        self._subscribers = {}

    def is_empty(self):
        return len(self._cache) == 0

    def add(self, obj):
        self._cache.append(obj)
        self._trigger_event(CacheEvent.ADD, obj)

    def remove(self, obj):
        self._cache.remove(obj)
        self._trigger_event(CacheEvent.REMOVE, obj)

    def clear(self):
        self._cache = []
        self._trigger_event(CacheEvent.CLEAR, None)

    def set_data(self, data):
        self._cache = copy.copy(data)
        self._trigger_event(CacheEvent.SET, self._cache)

    def _trigger_event(self, event: CacheEvent, param):
        subscribers = self._subscribers.setdefault(event, [])
        for sub in subscribers:
            sub(param)

    def add_event_listener(self, event: CacheEvent, fn):
        self._subscribers.setdefault(event, []).append(fn)

    def remove_event_listener(self, event: CacheEvent, fn):
        arr = self._subscribers.setdefault(event, [])
        self._subscribers[event] = [f for f in arr if f is not fn]

    def make_index(self, prop):
        index = CacheIndex(prop, self)
        index.on_set(self._cache)

        return index

    def get_data(self):
        return self._cache


class CacheIndex:
    def __init__(self, props, cache):
        self._cache = cache
        self.props = props if isinstance(props, list) else [props]

        self.index_map = {}

        cache.add_event_listener(CacheEvent.ADD, self.on_add)
        cache.add_event_listener(CacheEvent.REMOVE, self.on_remove)
        cache.add_event_listener(CacheEvent.CLEAR, self.on_clear)
        cache.add_event_listener(CacheEvent.SET, self.on_set)

    def __del__(self):
        cache = self._cache
        cache.remove_event_listener(CacheEvent.ADD, self.on_add)
        cache.remove_event_listener(CacheEvent.REMOVE, self.on_remove)
        cache.remove_event_listener(CacheEvent.CLEAR, self.on_clear)
        cache.remove_event_listener(CacheEvent.SET, self.on_set)

    def _add_with_prop(self, prop, obj):
        prop_val = getattr(obj, prop, None)

        if prop_val is not None:
            values = self.index_map.setdefault(prop_val, [])
            values.append(obj)

    def _remove_with_prop(self, prop, obj):
        prop_val = getattr(obj, prop, None)

        if prop_val is not None:
            values = self.index_map.setdefault(prop_val, [])
            values.remove(obj)

    def on_add(self, obj):
        for prop in self.props:
            self._add_with_prop(prop, obj)

    def on_remove(self, obj):
        for prop in self.props:
            self._remove_with_prop(prop, obj)

    def on_clear(self, param):
        self.index_map = {}

    def on_set(self, data):
        self.index_map = {}

        for entry in data:
            self.on_add(entry)

    def get(self, key, index=0):
        arr = self.index_map.setdefault(key, [])

        if len(arr) == 0:
            return None

        return arr[index]



from enum import Enum
from typing import Optional, List

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, kwargify


class CSVData(str):
    pass


class SecurityPolicyMatrixType(Enum):
    SECURITY_ZONE_MATRIX = "SECURITY_ZONE_MATRIX"


@propify
class SecurityPolicy(Jsonable):
    name: Optional[str] = prop(None)
    domain_id: Optional[str] = prop(None)
    domain_name: Optional[str] = prop(None)


@propify
class SecurityZoneMatrix(SecurityPolicy):
    description: Optional[str] = prop(None)
    type: Optional[SecurityPolicyMatrixType] = prop(None)


@propify
class SecurityPolicyInterface(Jsonable):
    name: str = prop(None)
    zones: List[str] = prop(factory=list, key="zones", flatify="zone")


@propify
class SecurityPolicyDeviceMapping(Jsonable):
    device_id: int = prop(None)
    affiliated_interfaces: List[SecurityPolicyInterface] = prop(factory=list)
    interfaces: List[SecurityPolicyInterface] = prop(
        factory=list, key="interfaces", flatify="interface"
    )


@propify
class ZoneUserAction(Jsonable):
    action: Optional[str] = prop(None)
    zone_id: Optional[str] = prop(None, key="zoneId")


@propify
class InterfaceUserMapping(Jsonable):
    interface_name: Optional[str] = prop(None)
    zones_user_actions: List[ZoneUserAction] = prop(factory=list)


@propify
class InterfacesManualMappings(Jsonable):
    interface_manual_mapping: List[InterfaceUserMapping] = prop(factory=list)
